import laspy
import argparse


class lascopy():
    def __init__(self):
        self.parse_args()
        self.copy_data()

    def parse_args(self):
        '''Set up the argparser to accept command line arguments.'''
        parser = argparse.ArgumentParser(
                description="""Accept in_file and out_file .LAS files, and 
                convert from in_file to out_file according to point_format
                and file_version.""")
        parser.add_argument('in_file', metavar='in_file', type=str, 
                            nargs='+', help='input file path')
        parser.add_argument('out_file', metavar='out_file', type=str, 
                            nargs='+', help='output file path')
        parser.add_argument('point_format', metavar='point_format', type=int, 
                            nargs='+', help='output point format (0-10)')
        parser.add_argument('file_version', metavar='file_version', type=str, 
                            nargs='+', help='output file version 1.0-1.4')
        parser.add_argument('-u', type=bool,
                help="Update the header histogram? (slow)", default=False)
        parser.add_argument('-b', type=bool,
                help="Attempt to preserve incompatable sub-byte sized data fields if present? (slow)", 
                default=False)

        self.args = parser.parse_args()

    def copy_data(self):
        ## Set several global flags to determine lascopy behavior
        UPDATE_HISTOGRAM= self.args.u
        PRESERVE = self.args.b

        ## Try to open in_file in read mode. 
        file_version = self.args.file_version[0]
        point_format = self.args.point_format[0]
        try:
            inFile = laspy.file.File(self.args.in_file[0], mode = "r")
        except Exception as error:
            print("There was an error reading in the input file: ")
            print(error)
            quit()

        ## Verify that the user has supplied a compatable point_format and file_version
        if not (point_format in range(11)):
            raise Exception("Invalid point format: %i" % point_format)
        if (point_format > 5 and file_version !="1.4"):
            raise Exception("Point formats 6-10 are only supported by version 1.4 files.")
        if (point_format > 3 and not (file_version in ["1.3", "1.4"])):
            raise Exception("Point format %i is not available for file version %s" %
                            (point_format, file_version))
        if (point_format >= 2 and not (file_version in ["1.2", "1.3", "1.4"])):
            raise Exception("Point format %i is not available for file version %s" % 
                            (point_format, file_version))

        ## Store depricated data for use later.
        old_file_version = inFile.header.version
        old_point_format = inFile.header.data_format_id


        ## Set global flag, which indicates whether the input and Output point formats 
        ## have compatable sub-byte fields. 
        SUB_BYTE_COMPATABLE = (old_point_format <= 5) == (point_format <= 5)

        ## Tell the user what we're doing. 
        print("Input File: " + self.args.in_file[0] + ", %i point records." % len(inFile))
        print("Output File: " + self.args.out_file[0])
        print("Converting from file version %s to version %s." %(old_file_version, file_version)) 
        print("Converting from point format %i to format %i."%(old_point_format, point_format))

        ## Warn the user if they chose potentially incompatable point formats and didn't specify -b=True
        if (not SUB_BYTE_COMPATABLE) and (not PRESERVE):
            print("""WARNING: The sub-byte sized fields differ between point formats 
                    %i and %i. By default this information will be lost. If you want 
                    laspy to try to preserve as much as possible, specify -b=True, though 
                    this may be quite slow depending on the size of the dataset.""" 
                    % (old_point_format, point_format))


        ## Build the new header, check if we need to do anythin special. This includes:
        ## 1. Check if there are extra dimensions defined which we need to copy.
        ## 2. Check if we need to get rid of some/all EVLRs for the output file_version
        ## 3. See if we need to re-map legacy fields for 1.4 files.
        try:
            new_header = inFile.header.copy()
            new_header.format = file_version 
            new_header.data_format_id = point_format

            old_data_rec_len = new_header.data_record_length
            old_std_rec_len = laspy.util.Format(old_point_format).rec_len
            diff =   old_data_rec_len - old_std_rec_len
            if (diff > 0):
                print("Extra Bytes Detected.")

            new_header.data_record_length = laspy.util.Format(point_format).rec_len + ((diff > 0)*diff)
            evlrs = inFile.header.evlrs
            if file_version != "1.4" and old_file_version == "1.4":
                print("Warning: input file has version 1.4, and output file does not. This may cause trunctation of header data.")
                new_header.point_return_count = inFile.header.legacy_point_return_count
                new_header.point_records_count = inFile.header.legacy_point_records_count
            if not (file_version in ["1.3", "1.4"]) and old_file_version in ["1.3", "1.4"]:
                print("Stripping any EVLRs")
                evlrs = []
            if (file_version == "1.3" and len(inFile.header.evlrs) > 1):
                print("Too many EVLRs for format 1.3, keeping the first one.")
                evlrs = inFile.header.evlrs[0]
            outFile = laspy.file.File(self.args.out_file[0], header = new_header, mode = "w", vlrs = inFile.header.vlrs, evlrs = evlrs)
            if outFile.point_format.rec_len != outFile.header.data_record_length:
                pass
        except Exception as error:
            print("There was an error instantiating the output file.")
            print(error)
            quit()

        ## Copy point dimensions. 
        try:
            for dimension in inFile.point_format.specs:
                if dimension.name in outFile.point_format.lookup:
                    ## Skip sub_byte field record bytes if incompatable
                    if (not SUB_BYTE_COMPATABLE and dimension.name in ("raw_classification", 
                        "classification_flags", "classification_byte", "flag_byte")):
                        continue
                    outFile.writer.set_dimension(dimension.name, inFile.reader.get_dimension(dimension.name))
                    print("Copying: " + dimension.name)
            ## Do we need to do anything special for sub-byte fields?
            if (not SUB_BYTE_COMPATABLE and PRESERVE):
                # Are we converting down or up?
                print("Attepmpting to copy sub-byte fields (this might take awhile)")
                up = old_point_format < point_format
                if up:
                    outFile.classification = inFile.classification
                    outFile.return_num = inFile.return_num
                    outFile.num_returns = inFile.num_returns
                    outFile.scan_dir_flag = inFile.scan_dir_flag
                    outFile.edge_flight_line = inFile.edge_flight_line
                    outFile.synthetic = inFile.synthetic
                    outFile.withheld = inFile.withheld
                    outFile.key_point = inFile.key_point
                else:
                    try:
                        outFile.classification = inFile.classification
                    except Exception as error:
                        print("Error, couldnt set classification.")
                        print(error)
                    try:
                        outFile.return_num = inFile.return_num
                    except Exception as error:
                        print("Error, coun't set return number.")
                        print(error)
                    try:
                        outFile.num_returns = inFile.num_returns
                    except:
                        print("Error, couldn't set number of returns.")
                    outFile.scan_dir_flag = inFile.scan_dir_flag
                    outFile.edge_flight_line = inFile.edge_flight_line
                    outFile.synthetic = inFile.synthetic
                    outFile.key_point = inFile.key_point
                    outFile.withheld = inFile.withheld


        except Exception as error:
            print("Error copying data.")
            print(error)
            quit()
        ## Close the files, and optionally update new file histogram.
        inFile.close()
        outFile.close(ignore_header_changes = not UPDATE_HISTOGRAM)



def main():
    lc = lascopy()
    print("Copy Complete")



from __future__ import absolute_import
from uuid import UUID
import unittest
import os
import time
import shutil
import laspy
from laspy.base import *
import laspy.file as File
import laspy.header as header


def flip_bit(x):
    return(1 * (x == 0))


class LasReaderTestCase(unittest.TestCase):
    simple = os.path.join(os.path.dirname(__file__), 'data', 'simple.las')
    tempfile = "junk.las"

    def setUp(self):
        really_copyfile(self.simple, self.tempfile)
        self.FileObject = File.File(self.tempfile)
        LasFile = self.FileObject
        self.X = list(LasFile.X)
        self.Y = list(LasFile.Y)
        self.Z = list(LasFile.Z)
        self.intensity = list(LasFile.intensity)
        self.flag_byte = list(LasFile.flag_byte)
        self.return_num = list(LasFile.return_num)
        self.num_returns = list(LasFile.num_returns)
        self.scan_dir_flag = list(LasFile.scan_dir_flag)
        self.edge_flight_line = list(LasFile.edge_flight_line)
        self.raw_classification = list(LasFile.raw_classification)
        self.classification = list(LasFile.classification)
        self.synthetic = list(LasFile.synthetic)
        self.key_point = list(LasFile.key_point)
        self.withheld = list(LasFile.key_point)
        self.scan_angle_rank = list(LasFile.scan_angle_rank)
        self.user_data = list(LasFile.user_data)
        self.pt_src_id = list(LasFile.pt_src_id)
        ## The following conditional code is redundant for 
        ## simple.las, which of course has only one pt. format.
        ## Perhaps find several other files?
        if LasFile._header.data_format_id in (1,2,3,4,5):
            self.gps_time = list(LasFile.gps_time)
        if LasFile._header.data_format_id in (2,3,5):
            self.red = list(LasFile.red)
            self.green = list(LasFile.green)
            self.blue = list(LasFile.blue)
        if LasFile._header.data_format_id in (4,5):
            self.wave_form_packet_Desc_index = list(LasFile.wave_packet_desc_index)
            self.byte_offset_to_waveform = list(LasFile.byte_offset_to_waveform)
            self.waveform_packet_size = list(LasFile.waveform_packet_size)
            self.return_point_waveform_loc = list(LasFile.return_point_waveform_loc)
            self.x_t = list(LasFile.x_t)
            self.y_t = list(LasFile.y_t )
            self.z_t = list(LasFile.z_t)

        self.p1 = LasFile.read(100)
        self.p2 = LasFile.read(976)

        self.idx1 = 100
        self.idx2 = 976


    def test_x(self):
        """Fetch and test X dimension"""
        self.assertEqual(self.p1.X , 63666106 , self.X[self.idx1])
        self.assertEqual(self.p2.X , 63714022 , self.X[self.idx2])

    def test_y(self):
        """Fetch and test Y dimension"""
        self.assertEqual(self.p1.Y , 84985413 , self.Y[self.idx1])
        self.assertEqual(self.p2.Y , 85318232 , self.Y[self.idx2])

    def test_z(self):
        """Fetch and test Z dimension"""
        self.assertEqual(self.p1.Z , 42490 , self.Z[self.idx1])
        self.assertEqual(self.p2.Z , 42306 , self.Z[self.idx2])


    def test_intensity(self):
        """Fetch and test intensity dimension"""
        self.assertEqual(self.p1.intensity , 233 ,
             self.intensity[self.idx1])
        self.assertEqual(self.p2.intensity , 1 ,
            self.intensity[self.idx2])


    def test_bit_flags(self):
        """Fetch and test the binary flags associated with flag_byte dimension"""
        self.assertEqual(self.p1.flag_byte , self.flag_byte[self.idx1])
        self.assertEqual(self.p2.flag_byte , self.flag_byte[self.idx2])
        self.assertEqual(self.p1.return_num ,
            self.return_num[self.idx1] , 1)
        self.assertEqual(self.p2.return_num ,
            self.return_num[self.idx2] , 2)
        self.assertEqual(self.p1.num_returns ,
            self.num_returns[self.idx1] , 1)
        self.assertEqual(self.p2.num_returns ,
            self.num_returns[self.idx2] , 2)
        self.assertEqual(self.p1.scan_dir_flag ,
            self.scan_dir_flag[self.idx1] , 0)
        self.assertEqual(self.p2.scan_dir_flag ,
            self.scan_dir_flag[self.idx2] , 1)
        self.assertEqual(self.p1.edge_flight_line ,
            self.edge_flight_line[self.idx1] , 0)
        self.assertEqual(self.p2.edge_flight_line ,
            self.edge_flight_line[self.idx2] , 0)


    def test_scan_angle_rank(self):
        """Fetch and test scan_angle_rank dimension"""
        self.assertEqual(self.p1.scan_angle_rank , 2 ,
            self.scan_angle_rank[self.idx1])
        self.assertEqual(self.p2.scan_angle_rank , 12 ,
            self.scan_angle_rank[self.idx2])


    def test_raw_classification(self):
        """Fetch and test the dimension of raw_classification bytes"""
        self.assertEqual(self.p1.raw_classification , 1 ,
                self.raw_classification[self.idx1])
        self.assertEqual(self.p2.raw_classification , 2 ,
                 self.raw_classification[self.idx2])


    def test_ptstrcid(self):
        """Fetch and test pt_src_id dimension"""
        self.assertEqual(self.p1.pt_src_id , 7328 ,
            self.pt_src_id[self.idx1])
        self.assertEqual(self.p2.pt_src_id , 7334 ,
            self.pt_src_id[self.idx2])


    def test_GPSTime(self):
        """Fetch and test gps_time dimension"""
        self.assertTrue(self.p1.gps_time - 2*246504.221932
                    + self.gps_time[self.idx1] < 0.00001)
        self.assertTrue(self.p2.gps_time - 2*249774.658254
                    + self.gps_time[self.idx2] < 0.00001)


    def test_red(self):
        """Fetch and test red dimension"""
        self.assertEqual(self.p1.red , 92 , self.red[self.idx1])
        self.assertEqual(self.p2.red , 94 , self.red[self.idx2])


    def test_green(self):
        """Fetch and test green dimension"""
        self.assertEqual(self.p1.green , 100 , self.green[self.idx1])
        self.assertEqual(self.p2.green , 84 , self.green[self.idx2])


    def test_blue(self):
        """Fetch and test blue dimension"""
        self.assertEqual(self.p1.blue , 110 , self.blue[self.idx1])
        self.assertEqual(self.p2.blue , 94 , self.blue[self.idx2])

    def test_iterator_and_slicing(self):
        """Test iteraton and slicing over File objects"""
        k = 0
        LasFile = self.FileObject
        for pt1 in LasFile:
            pt2 = LasFile[k]
            pt1.make_nice()
            pt2.make_nice()
            self.assertEqual(pt1.X, pt2.X)
            self.assertEqual(pt1.Y, pt2.Y)
            self.assertEqual(pt1.Z, pt2.Z)
            self.assertEqual(pt1.gps_time, pt2.gps_time)
            k += 1
        with self.assertRaises(Exception):
            LasFile[10000]

    def tearDown(self):
        self.FileObject.close()
        really_remove(self.tempfile)


class LasWriterTestCase(unittest.TestCase):
    simple = os.path.join(os.path.dirname(__file__), 'data', 'simple.las')
    tempfile = 'writer.las'
    output_tempfile = 'writer_output.las'
    def setUp(self):
        really_copyfile(self.simple, self.tempfile)
        self.FileObject = File.File(self.tempfile, mode = "rw")

    def test_x(self):
        """Writing and testing X dimenson"""
        x = [i + 1 for i in self.FileObject.X]
        self.FileObject.X = x
        x2 = self.FileObject.get_x()
        self.assertTrue((list(x) == list(x2)))
        scaled_x = self.FileObject.x
        self.FileObject.x = scaled_x
        self.assertTrue(all(scaled_x == self.FileObject.x))

    def test_y(self):
        """Writing and testing Y dimension"""
        y = [i + 1 for i in self.FileObject.Y]
        self.FileObject.Y = y
        y2 = self.FileObject.get_y()
        self.assertTrue((list(y) == list(y2)))
        scaled_y = self.FileObject.y
        self.FileObject.y = scaled_y
        self.assertTrue(all(scaled_y == self.FileObject.y))

    def test_z(self):
        """Writing and testing Z dimension"""
        z = [i + 1 for i in self.FileObject.Z ]
        self.FileObject.Z = z
        z2 = self.FileObject.get_z()
        self.assertTrue((list(z) == list(z2)))
        scaled_z = self.FileObject.z
        self.FileObject.z = scaled_z
        self.assertTrue(all(scaled_z == self.FileObject.z))

    def test_intensity(self):
        """Writing and testing intensity dimension"""
        i = [i + 1 for i in self.FileObject.intensity]
        self.FileObject.intensity = i
        i2 = self.FileObject.intensity
        self.assertTrue((i == list(i2)))
    def test_return_num(self):
        """Writing and testing return_num dimension"""
        rn = [i + 1 for i in self.FileObject.return_num]
        self.FileObject.return_num = rn
        rn2 = self.FileObject.get_return_num()
        self.assertTrue((rn == list(rn2)))
    def test_overflow_return_num(self):
        """Testing overflow handling"""
        rn = [x + 100000 for x in self.FileObject.return_num]
        with self.assertRaises(Exception):
            self.FileObject.return_num = rn
    def test_num_returns(self):
        """Writing and testing num_returns dimension"""
        nr = [i + 1 for i in self.FileObject.num_returns]
        self.FileObject.num_returns = nr
        nr2 = self.FileObject.get_num_returns()
        self.assertTrue((nr == list(nr2)))
    def test_scan_dir_flag(self):
        """Writing and testing scan_dir_flag dimension"""
        sdf = [flip_bit(x) for x in self.FileObject.scan_dir_flag]
        self.FileObject.scan_dir_flag = sdf
        sdf2 = self.FileObject.get_scan_dir_flag()
        self.assertTrue((sdf == list(sdf2)))
    def test_edge_flight_line(self):
        """Writing and testing edge_flight_line dimension"""
        efl = [flip_bit(x) for x in self.FileObject.edge_flight_line]
        self.FileObject.edge_flight_line = efl
        efl2 = self.FileObject.get_edge_flight_line()
        self.assertTrue((efl == list(efl2)))
    def test_classification(self):
        """Writing and testing classification byte."""
        c1 = [x + 1 for x in self.FileObject.classification]
        self.FileObject.classification = c1
        c2 = [x for x in self.FileObject.get_classification()]
        self.assertTrue((c1 == c2))
    def test_synthetic(self):
        """Writing and testing synthetic"""
        s1 = [flip_bit(x) for x in self.FileObject.synthetic]
        self.FileObject.synthetic = s1
        s2 = self.FileObject.get_synthetic()
        self.assertTrue((s1 == list(s2)))
    def test_key_point(self):
        """Writing and testing key point"""
        k1 = [flip_bit(x) for x in self.FileObject.key_point]
        self.FileObject.key_point = k1
        k2 = self.FileObject.get_key_point()
        self.assertTrue((k1 == list(k2)))
    def test_withheld(self):
        """Writing and testing withheld"""
        w1 = [flip_bit(x) for x in self.FileObject.withheld]
        self.FileObject.withheld = w1
        w2 = self.FileObject.get_withheld()
        self.assertTrue((w1 == list(w2)))
    def test_scan_angle_rank(self):
        """Writing and testing scan angle rank"""
        ar1 = [i-1 for i in self.FileObject.scan_angle_rank]
        ar1 = [max(0, x) for x in ar1]
        self.FileObject.scan_angle_rank = ar1
        ar2 = self.FileObject.get_scan_angle_rank()
        self.assertTrue((ar1 == list(ar2)))
    def test_user_data(self):
        """Writing and testing user data"""
        ud1 = [i+1 for i in self.FileObject.user_data]
        self.FileObject.user_data = ud1
        ud2 = self.FileObject.get_user_data()
        self.assertTrue((ud1 == list(ud2)))
    def test_pt_src_id(self):
        """Writing and testing point source id"""
        p1 = [i+1 for i in self.FileObject.user_data]
        self.FileObject.user_data = p1
        p2 = self.FileObject.get_user_data()
        self.assertTrue((p1 == list(p2)))
    def test_gps_time(self):
        """Writing and testing gps time"""
        g1 = [i+1 for i in self.FileObject.gps_time]
        self.FileObject.gps_time = g1
        g2 = self.FileObject.get_gps_time()
        self.assertTrue((g1 == list(g2)))
    def test_red(self):
        """Writing and testing red"""
        r1 = [i+1 for i in self.FileObject.red]
        self.FileObject.red = r1
        r2 = self.FileObject.get_red()
        self.assertTrue((r1 == list(r2)))
    def test_green(self):
        """Writing and testing green"""
        g1 = [i+1 for i in self.FileObject.green]
        self.FileObject.green = g1
        g2 = self.FileObject.get_green()
        self.assertTrue((g1 == list(g2)))
    def test_blue(self):
        """Writing and testing blue"""
        b1 =[i+1 for i in  self.FileObject.blue]
        self.FileObject.blue = b1
        b2 = self.FileObject.get_blue()
        self.assertTrue((b1 == list(b2)))
    def test_vlr_defined_dimensions2(self):
        """Testing VLR defined dimension API"""
        File2 = File.File(self.output_tempfile, mode = "w", header = self.FileObject.header)
        File2.define_new_dimension("test_dimension", 5, "This is a test.")
        File2.X = self.FileObject.X
        self.assertTrue(File2.test_dimension[500] == 0)
        File2.close(ignore_header_changes = True)
    #def test_wave_pkt_descp_idx(self):
    #    w1 = self.FileObject.wave_packet_descp_idx + 1
    #    self.FileObject.wave_packet_descp_idx = w1
    #    w2 = self.FileObject.get_wave_packet_descp_idx()
    #    self.assertTrue((w1 = w2))
    #def test_byte_offset(self):
    #    b1 = self.FileObject.byte_offset_to_waveform_data + 1
    #    self.FileObject.byte_offset_to_waveform_data = b1
    #    b2 = self.FileObject.get_byte_offset_to_waveform_data
    #    self.assertTrue((b1 == b2))
    #def test_wavefm_pkt_size(self):
    #    w1 = self.FileObject.waveform_pkt_size + 1
    #    self.FileObject.waveform_pkt_size = w1
    #    w2 = self.FileObject.get_waveform_pkt_size()
    #    self.assertTrue((w1 == w2))
    #def test_return_pt_wavefm_loc(self):
    #    w1 = self.FileObject.return_point_waveform_loc + 1
    #    self.FileObject.return_point_waveform_loc = w1
    #    w2 = self.FileObject.get_return_point_waveform_loc
    #    self.assertTrue((w1 == w2))
    #def test_x_t(self):
    #    x1 = self.FileObject.x_t + 1
    #    self.FileObject.x_t = x1
    #    x2 = self.FileObject.get_x_t()
    #    self.assertTrue((x1 == x2))
    #def test_y_t(self):
    #    y1 = self.FileObject.y_t + 1
    #    self.FileObject.y_t = y1
    #    y2 = self.FileObject.get_y_t()
    #    self.assertTrue((y1 == y2))
    #def test_z_t(self):
    #    z1 = self.FileObject.z_t + 1
    #    self.FileObject.z_t = z
    #    z2 = self.FileObject.get_z_t()
    #    self.assertTrue((z1 == z2))    
    def tearDown(self):
        self.FileObject.close()
        really_remove(self.tempfile)
        really_remove(self.output_tempfile)

class LasHeaderWriterTestCase(unittest.TestCase):
    simple = os.path.join(os.path.dirname(__file__), 'data', 'simple.las')
    simple14 = os.path.join(os.path.dirname(__file__), 'data', 'simple1_4.las')
    tempfile = os.path.abspath('headerwriter.las')
    tempfile2 = os.path.abspath('headerwriter2.las')

    def setUp(self):
        really_copyfile(self.simple, self.tempfile)
        self.FileObject = File.File(self.tempfile, mode = "rw")

    def test_file_src(self):
        """Testing file_src"""
        f1 = self.FileObject.header.filesource_id + 1
        self.FileObject.header.filesource_id = f1
        f2 = self.FileObject.header.get_filesourceid()
        self.assertTrue(f1 == f2)
    def test_uuid(self):
        """Testing uuid"""
        guid = self.FileObject.header.guid
        guid2 = self.FileObject.header.project_id
        self.assertEqual(guid, guid2)
        newGuid = UUID(bytes=b"1"*16)
        self.FileObject.header.guid = newGuid
        newGuid2 = self.FileObject.header.get_guid()
        self.assertEqual(newGuid, newGuid2)
    def test_glob_encode(self):
        """Testing Global Encoding"""
        g1 = self.FileObject.header.global_encoding + 1
        self.FileObject.header.global_encoding = g1
        g2 = self.FileObject.header.get_global_encoding()
        self.assertTrue(g1 == g2)
    def test_versions(self):
        """Testing Versions"""
        v1 = self.FileObject.header.major_version
        self.assertEqual(v1, 1)
        with self.assertRaises(laspy.util.LaspyException):
            self.FileObject.header.major_version = 2
    def test_system_id(self):
        """Testing System ID"""
        sys1 = self.FileObject.header.system_id
        sys1 = "1234567891" + sys1[10:]
        self.FileObject.header.system_id = sys1
        sys2 = self.FileObject.header.get_systemid()
        self.assertEqual(sys1, sys2)
    def test_software_id(self):
        """"Testing Software ID"""
        s1 = self.FileObject.header.software_id
        s1 = "1234567" + s1[7:]
        self.FileObject.header.software_id = s1
        s2 = self.FileObject.header.get_softwareid()
        self.assertEqual(s1, s2)
        with self.assertRaises(laspy.util.LaspyException):
            self.FileObject.header.software_id = "123"
        with self.assertRaises(laspy.util.LaspyException):
            self.FileObject.header.software_id = "1" * 100
    def test_padding(self):
        """Testing Padding"""
        x1 = list(self.FileObject.X)
        self.FileObject.header.set_padding(10)
        self.FileObject.header.set_padding(1000)
        x2 = list(self.FileObject.X)
        self.assertTrue((list(x1) == list(x2)))
    def test_data_offset(self):
        """Testing data offset"""
        x1 = list(self.FileObject.X )
        self.FileObject.header.data_offset = 400
        self.assertEqual(self.FileObject.header.get_dataoffset(), 400)
        x2 = list(self.FileObject.X)
        self.assertTrue((list(x1) == list(x2)))
    def test_date(self):
        """Testing Date"""
        d1 = self.FileObject.header.date
        self.assertTrue(d1 == None)
        from datetime import datetime
        d2 = datetime(2007,12,10)
        self.FileObject.header.date = d2
        d3 = self.FileObject.header.get_date()
        self.assertEqual(d2, d3)

    def test_point_recs_by_return(self):
        """Testing point records by return"""
        r1 = [x + 1 for x in self.FileObject.header.point_return_count]
        self.FileObject.header.point_return_count = r1
        r2 = self.FileObject.header.get_pointrecordsbyreturncount()
        self.assertTrue(r1 == r2)
    def test_min_max_update(self):
        """Testing the update min/max function"""
        x = list(self.FileObject.X)
        x[0] = max(x) + 1
        y = list(self.FileObject.Y)
        y[0] = max(y) + 1
        z = list(self.FileObject.Z)
        z[0] = max(z) + 1
        self.FileObject.X = x
        self.FileObject.Y = y
        self.FileObject.Z = z
        self.FileObject.header.update_min_max()
        file_max = self.FileObject.header.max
        self.assertTrue(file_max == [self.FileObject.x[0], self.FileObject.y[0], self.FileObject.z[0]])
    def test_histogram(self):
        """Testing the update_histogram functon"""
        h1 = self.FileObject.header.point_return_count
        self.FileObject.header.update_histogram()
        h2 = self.FileObject.header.point_return_count
        self.assertEqual(h1, h2)
    def test_offset(self):
        """Testing offset"""
        o1 = self.FileObject.header.offset
        o1[0] += 1
        self.FileObject.header.offset = o1
        o2 = self.FileObject.header.get_offset()
        self.assertTrue(o1 == o2)
    def test_scale(self):
        """Testing Scale"""
        s1 = self.FileObject.header.scale
        s1[0] += 1
        self.FileObject.header.scale = s1
        s2 = list(self.FileObject.header.get_scale())
        self.assertTrue(s1 == s2)

    def test_vlr_parsing_api(self):
        """Testing VLR body parsing api"""
        really_copyfile(self.simple14, self.tempfile2)
        VLRFile = File.File(self.tempfile2, mode = "rw")
        vlr0 = VLRFile.header.vlrs[0]
        pb = vlr0.parsed_body
        vlr0.parsed_body = pb[:-1] + (pb[-1] + 1,)
        pb = vlr0.parsed_body
        vlr0.pack_data()
        VLRFile.header.save_vlrs()
        VLRFile.close()

        VLRFile = File.File(self.tempfile2, mode = "rw")
        self.assertTrue(pb == VLRFile.header.vlrs[0].parsed_body)
        VLRFile.close()

    def tearDown(self):
        self.FileObject.close()
        really_remove(self.tempfile)
        really_remove(self.tempfile2)

class LasWriteModeTestCase(unittest.TestCase):
    simple = os.path.join(os.path.dirname(__file__), 'data', 'simple.las')
    tempfile = 'write-mode.las'
    output_tempfile = 'write-mode-output.las'
    def setUp(self):
        really_copyfile(self.simple, self.tempfile)
        self.File1 = File.File(self.tempfile, mode = "r")

    def test_using_barebones_header(self):
        """Testing file creation using barebones header"""
        header_object = header.Header()

        File2 = File.File(self.output_tempfile, mode = "w",
                            header = header_object)
        self.assertTrue(File2.header.version == "1.2")

        X = list(self.File1.X)
        Y = list(self.File1.Y)
        Z = list(self.File1.Z)
        File2.X = X
        File2.Y = Y
        File2.Z = Z
        self.assertTrue((list(X) == list(File2.get_x())))
        self.assertTrue((list(Y) == list(File2.get_y())))
        self.assertTrue((list(Z) == list(File2.get_z())))
        File2.close(ignore_header_changes = True)

    def test_using_existing_header(self):
        """Test file creation using an existing header"""
        header_object = self.File1.header
        File2 = File.File(self.output_tempfile, mode = "w",
                            header = header_object)
        X = list(self.File1.X)
        Y = list(self.File1.Y)
        Z = list(self.File1.Z)
        File2.Z = Z
        File2.Y = Y
        File2.X = X
        self.assertTrue((list(X) == list(File2.get_x())))
        self.assertTrue((list(Y) == list(File2.get_y())))
        self.assertTrue((list(Z) == list(File2.get_z())))
        File2.close(ignore_header_changes = True)

    def test_format_change_and_extra_bytes(self):
        """Testing format change and extra_bytes"""
        File1 = self.File1
        new_header = File1.header.copy()
        new_header.format = 1.2
        new_header.data_format_id = 0
        new_header.data_record_length = 50


        File2 = File.File(self.output_tempfile, mode = "w",
                            header = new_header)
        for dim in File1.point_format:
            in_dim = File1.reader.get_dimension(dim.name)
            if dim.name in File2.point_format.lookup:
                File2.writer.set_dimension(dim.name, in_dim)
        File2.extra_bytes = [b"Test"] * len(File2)
        self.assertTrue(b"Test" in bytes(File2.get_extra_bytes()[14]))
        File2.close(ignore_header_changes = True)
    def tearDown(self):
        self.File1.close()
        really_remove(self.tempfile)
        really_remove(self.output_tempfile)

class LasV_13TestCase(unittest.TestCase):
    simple = os.path.join(os.path.dirname(__file__), 'data', 'simple1_3.las')
    tempfile = 'v13.las'
    output_tempfile = 'v13-output.las'
    def setUp(self):
        really_copyfile(self.simple, self.tempfile)
        self.File1 = File.File(self.tempfile, mode = "rw")

    def test_glob_encode(self):
        """Testing v1.3 Global Encoding"""
        old = self.File1.header.gps_time_type
        self.assertEqual(old, '0')
        self.File1.header.gps_time_type = '1'
        self.assertEqual(self.File1.header.get_gps_time_type(), '1')

        File2 = File.File(self.output_tempfile, mode = "w", header = self.File1.header)
        self.assertEqual(self.File1.header.waveform_data_packets_internal,
                        File2.header.waveform_data_packets_internal)
        self.assertEqual(self.File1.header.waveform_data_packets_external,
                        File2.header.waveform_data_packets_external)
        self.assertEqual(self.File1.header.synthetic_return_num,
                        File2.header.synthetic_return_num)
        File2.close(ignore_header_changes = True)

    def test_evlr(self):
        """Testing v1.3 EVLR support."""
        File2 = File.File(self.output_tempfile, mode = "w", header = self.File1.header)
        self.assertEqual(self.File1.header.evlrs[0].to_byte_string(),
                        File2.header.evlrs[0].to_byte_string())
        File2.close(ignore_header_changes = True)
    def test_wave_packet_desc_index(self):
        """Testing wave_packet_desc_index."""
        test1 = self.File1.wave_packet_desc_index
        File2 = File.File(self.output_tempfile, mode = "w", header = self.File1.header)
        File2.points = self.File1.points
        self.assertTrue(all(test1 == File2.wave_packet_desc_index))
        File2.wave_packet_desc_index += 1
        self.assertTrue(all(test1 != File2.wave_packet_desc_index))
        File2.close(ignore_header_changes = True)

    def test_byte_offset_to_waveform_data(self):
        """Testing byte_offset_to_waveform_data"""
        test1 = self.File1.byte_offset_to_waveform_data
        File2 = File.File(self.output_tempfile, mode = "w", header = self.File1.header)
        File2.points = self.File1.points
        self.assertTrue(all(test1 == File2.byte_offset_to_waveform_data))
        File2.byte_offset_to_waveform_data += 1
        self.assertTrue(all(test1 != File2.byte_offset_to_waveform_data))
        File2.close(ignore_header_changes = True)
    def test_waveform_packet_size(self):
        """Testing waveform_packet_size"""
        test1 = self.File1.waveform_packet_size
        File2 = File.File(self.output_tempfile, mode = "w", header = self.File1.header)
        File2.points = self.File1.points
        self.assertTrue(all(test1 == File2.waveform_packet_size))
        File2.waveform_packet_size += 1
        self.assertTrue(all(test1 != File2.waveform_packet_size))
        File2.close(ignore_header_changes = True)
    def test_return_point_waveform_loc(self):
        """Testing return_point_waveform_loc"""
        test1 = self.File1.return_point_waveform_loc
        File2 = File.File(self.output_tempfile, mode = "w", header = self.File1.header)
        File2.points = self.File1.points
        self.assertTrue(all(test1 == File2.return_point_waveform_loc))
        File2.return_point_waveform_loc += 1
        self.assertTrue(all(test1 != File2.return_point_waveform_loc))
        File2.close(ignore_header_changes = True)
    def test_x_t(self):
        """Testing x_t"""
        test1 = self.File1.x_t
        File2 = File.File(self.output_tempfile, mode = "w", header = self.File1.header)
        File2.points = self.File1.points
        self.assertTrue(all(test1 == File2.x_t))
        File2.x_t += 1
        self.assertTrue(all(test1 != File2.x_t))
        File2.close(ignore_header_changes = True)
    def test_y_t(self):
        """Testing y_t"""
        test1 = self.File1.y_t
        File2 = File.File(self.output_tempfile, mode = "w", header = self.File1.header)
        File2.points = self.File1.points
        self.assertTrue(all(test1 == File2.y_t))
        File2.y_t += 1
        self.assertTrue(all(test1 != File2.y_t))
        File2.close(ignore_header_changes = True)
    def test_z_t(self):
        """Testing z_t"""
        test1 = self.File1.z_t
        File2 = File.File(self.output_tempfile, mode = "w", header = self.File1.header)
        File2.points = self.File1.points
        self.assertTrue(all(test1 == File2.z_t))
        File2.z_t += 1
        self.assertTrue(all(test1 != File2.z_t))
        File2.close(ignore_header_changes = True)

    def tearDown(self):
        self.File1.close()
        really_remove(self.tempfile)
        really_remove(self.output_tempfile)


class LasV_14TestCase(unittest.TestCase):
    simple = os.path.join(os.path.dirname(__file__), 'data', 'simple1_4.las')
    tempfile = 'v14.las'
    output_tempfile = 'v14-output.las'
    def setUp(self):
        really_copyfile(self.simple, self.tempfile)
        self.File1 = File.File(self.tempfile, mode = "rw")

    def test_glob_encode(self):
        """Testing v1.4 Global Encoding"""
        old = self.File1.header.gps_time_type
        self.assertEqual(old, '0')
        self.File1.header.gps_time_type = '1'
        self.assertEqual(self.File1.header.get_gps_time_type(), '1')

        File2 = File.File(self.output_tempfile, mode = "w", header = self.File1.header)
        self.assertEqual(self.File1.header.waveform_data_packets_internal,
                        File2.header.waveform_data_packets_internal)
        self.assertEqual(self.File1.header.waveform_data_packets_external,
                        File2.header.waveform_data_packets_external)
        self.assertEqual(self.File1.header.synthetic_return_num,
                        File2.header.synthetic_return_num)
        File2.close(ignore_header_changes = True)

    def test_glob_encode_bits(self):
        b1 = self.File1.header.gps_time_type
        b2 = self.File1.header.waveform_data_packets_internal
        b3 = self.File1.header.waveform_data_packets_external
        b4 = self.File1.header.synthetic_return_num
        b5 = self.File1.header.wkt

        bf1 = 1 - int(b1)
        bf2 = 1 - int(b2)
        bf3 = 1 - int(b3)
        bf4 = 1 - int(b4)
        bf5 = 1 - int(b5)

        self.File1.header.gps_time_type = bf1
        self.File1.header.waveform_data_packets_internal = bf2
        self.File1.header.waveform_data_packets_external = bf3
        self.File1.header.synthetic_return_num = bf4
        self.File1.header.wkt = bf5

        self.assertEqual(self.File1.header.gps_time_type, str(bf1))
        self.assertEqual(self.File1.header.waveform_data_packets_internal, str(bf2))
        self.assertEqual(self.File1.header.waveform_data_packets_external, str(bf3))
        self.assertEqual(self.File1.header.synthetic_return_num, str(bf4))
        self.assertEqual(self.File1.header.wkt, str(bf5))

    def test_evlr(self):
        """ Testing v1.4 EVLR support"""
        File2 = File.File(self.output_tempfile, mode = "w", header = self.File1.header)
        self.assertEqual(self.File1.header.evlrs[0].to_byte_string(),
                        File2.header.evlrs[0].to_byte_string())
        File2.points = self.File1.points
        outevlrs = []
        [outevlrs.append(File2.header.evlrs[0]) for i in xrange(50)]
        File2.header.evlrs = outevlrs
        File2.close()
        File2 = File.File(self.output_tempfile, mode = "r")
        self.assertEqual(len(File2.header.evlrs), 50)
        File2.close(ignore_header_changes = True)

    def test_classification_variables(self):
        """Testing v1.4 classification support"""
        classification = self.File1.classification
        classification_flags = self.File1.classification_flags
        scanner_channel = self.File1.scanner_channel
        scan_dir_flag = self.File1.scan_dir_flag
        edge_flight_line = self.File1.edge_flight_line

        return_num = self.File1.return_num
        num_returns = self.File1.num_returns
        File2 = File.File(self.output_tempfile, mode = "w", header = self.File1.header)
        File2.classification = classification
        File2.classification_flags = classification_flags
        File2.scan_dir_flag = scan_dir_flag
        File2.scanner_channel = scanner_channel
        File2.edge_flight_line = edge_flight_line
        File2.return_num = return_num
        File2.num_returns = num_returns

        self.assertTrue(all(num_returns == File2.get_num_returns()))
        self.assertTrue(all(return_num == File2.get_return_num()))
        self.assertTrue(all(edge_flight_line == File2.get_edge_flight_line()))
        self.assertTrue(all(scan_dir_flag ==File2.get_scan_dir_flag() ))
        self.assertTrue(all(classification_flags == File2.get_classification_flags()))
        self.assertTrue(all(classification == File2.get_classification()))
        self.assertTrue(all(scanner_channel == File2.get_scanner_channel()))
        File2.close(ignore_header_changes = True)


    def test_vlr_defined_dimensions(self):
        """Testingi multiple v1.4 VLR defined dimensions (LL API)"""
        new_header = self.File1.header.copy()
        # Test basic numeric dimension
        new_dim_record1 = header.ExtraBytesStruct(name = "test dimension 1234", data_type = 5)
        # Test string dimension (len 3)
        new_dim_record2 = header.ExtraBytesStruct(name = "test dimension 5678", data_type = 22)
        # Test integer array dimension (len 3)
        new_dim_record3 = header.ExtraBytesStruct(name = "test dimension 9", data_type =  26)
        new_VLR_rec = header.VLR(user_id = "LASF_Spec", record_id = 4,
            VLR_body = (new_dim_record1.to_byte_string() + new_dim_record2.to_byte_string() + new_dim_record3.to_byte_string()))
        new_header.data_record_length += (19)
        File2 = File.File(self.output_tempfile, mode = "w", header = new_header, 
                vlrs = [new_VLR_rec], evlrs = self.File1.header.evlrs)

        File2.X = self.File1.X

        File2._writer.set_dimension("test_dimension_1234", [4]*len(self.File1))
        File2._writer.set_dimension("test_dimension_5678", ["AAA"]*len(self.File1))
        File2._writer.set_dimension("test_dimension_9", [[1,2,3]]*len(self.File1))
        self.assertTrue(all(np.array([4]*len(self.File1)) == File2.test_dimension_1234))
        self.assertTrue(all(np.array([b"AAA"]*len(self.File1)) == File2.test_dimension_5678))
        self.assertEqual(list(File2.test_dimension_9[100]), [1,2,3])
        File2.close(ignore_header_changes = True)


    def test_vlr_defined_dimension_dtypes(self):
        """Testing v1.4 VLR defined dimension data types (LL API)"""
        from laspy.util import edim_fmt_dict, fmtLen, LEfmt,defaults,defaults_test

        for dfList in [defaults, defaults_test]:

            for i in range(1,31):
                print("...data format: "+ str(i))
                # Create a new header
                new_header = self.File1.header.copy()
                # Create new dimension
                dimname = "test_dimension_" + str(i)
                new_dimension = header.ExtraBytesStruct(
                    name = dimname, data_type = i) 
                # Collect bytes for new dimension specification 
                new_dim_raw = new_dimension.to_byte_string() 
                # Create a VLR defining our new dim
                new_VLR_rec = header.VLR(user_id = "LASF_Spec", record_id = 4,
                    description = "Testing Extra Bytes.",
                    VLR_body = (new_dim_raw))

                # Figure out how much we need to pad the point records
                new_dim_fmt = edim_fmt_dict[i]
                new_dim_num = new_dim_fmt[1]
                new_dim_bytelen = fmtLen[LEfmt[new_dim_fmt[0]]]
                new_total_bytes = new_dim_bytelen*new_dim_num 

                new_header.data_record_length += (new_total_bytes)

                File2 = File.File(self.output_tempfile, mode = "w", header = new_header, 
                    vlrs = [new_VLR_rec], evlrs = self.File1.header.evlrs)
    
                File2.X = self.File1.X
    
                dim_default = dfList[LEfmt[edim_fmt_dict[i][0]]]
                is_str = type(dim_default) == type(" ")
                if new_dim_num == 1:
                    new_dim_val = [dim_default]*len(self.File1)
                elif is_str:
                    new_dim_val = [dim_default*new_dim_num]*len(self.File1)
                else: 
                    new_dim_val = [[dim_default]*new_dim_num]*len(self.File1)

                File2._writer.set_dimension(dimname, new_dim_val)
                current_dim_val = File2._writer.get_dimension(dimname)

                if (is_str):
                    self.assertEqual([x.encode() for x in new_dim_val],
                            current_dim_val.tolist(), msg = "Problem with data format " + str(i))
                else:
                    self.assertEqual(current_dim_val.tolist(), new_dim_val,
                        msg = "Problem with data format " + str(i))
                
                File2.close(ignore_header_changes = True)
    
    
    def test_vlr_defined_dimensions2(self):
        """Testing VLR defined dimensions (HL API)"""
        File2 = File.File(self.output_tempfile, mode = "w", header = self.File1.header)
        File2.define_new_dimension("test_dimension", 5, "This is a test.")
        File2.define_new_dimension("test_dimension2", 5, "This is a test.")

        File2.X = self.File1.X
        self.assertEqual(File2.test_dimension[500], 0)
        self.assertEqual(File2.test_dimension2[123], 0)
        File2.close(ignore_header_changes = True)

    def test_vlr_defined_dimensions3(self):
        """Testing VLR defined dimensions (HL API)"""
        File2 = File.File(self.output_tempfile, mode = "w", header = self.File1.header)
        File2.define_new_dimension("Test_Dimension", 5, "This is a test.")

        File2.X = self.File1.X
        self.assertEqual(File2.Test_Dimension[500], 0)
        File2.close(ignore_header_changes = True)



    def tearDown(self):
        self.File1.close()
        really_remove(self.tempfile)
        really_remove(self.output_tempfile)


class LasLazReaderTestCase(unittest.TestCase):
    def setUp(self):
        pass

    def test_read_data(self):
        """Testing ability to read laz files."""
        simple = os.path.join(os.path.dirname(__file__), 'data', 'simple.laz')
        File1 = File.File(simple, mode = "r")
        self.assertTrue(len(File1.X) == 1065)
        File1.close()

    def tearDown(self):
        pass


class LazNoPointMapTestCase(unittest.TestCase):

    def test_open_header_only(self):
        """Test opening in mode r- (do not uncompress)"""
        simple = os.path.join(os.path.dirname(__file__), "data", "simple.laz")
        plas = File.File(simple, mode="r-")
        self.assertTrue(plas.header.records_count == 1065)
        plas.close()

    def test_open_invalid(self):
        simple = os.path.join(os.path.dirname(__file__), "data", "simple.laz")
        with self.assertRaises(Exception):
            plas = File.File(simple, mode="rr")

class LazReaderTestCase(unittest.TestCase):
    simple_z = os.path.join(os.path.dirname(__file__), 'data', 'simple.laz')
    simple_s = os.path.join(os.path.dirname(__file__), 'data', 'simple.las')

    def test_open(self):
        """Test opening LAZ file with lazperf"""
        plas = File.File(self.simple_z, mode="r")
        self.assertTrue(plas.header.records_count == 1065)
        plas.close()

    def test_x(self):
        """Fetch and test X, Y, Z dimensions"""
        compressed = File.File(self.simple_z, mode="r")
        uncompressed = File.File(self.simple_s, mode="r")
        X_z = list(compressed.X)
        Y_z = list(compressed.Y)
        Z_z = list(compressed.Z)
        X_s = list(uncompressed.X)
        Y_s = list(uncompressed.Y)
        Z_s = list(uncompressed.Z)
        self.assertTrue((list(X_z) == list(X_s)))
        self.assertTrue((list(Y_z) == list(Y_s)))
        self.assertTrue((list(Z_z) == list(Z_s)))

def test_laspy():
    reader = unittest.TestLoader().loadTestsFromTestCase(LasReaderTestCase)
    writer = unittest.TestLoader().loadTestsFromTestCase(LasWriterTestCase)
    header_writer = unittest.TestLoader().loadTestsFromTestCase(LasHeaderWriterTestCase)
    write_mode = unittest.TestLoader().loadTestsFromTestCase(LasWriteModeTestCase)
    las13 = unittest.TestLoader().loadTestsFromTestCase(LasV_13TestCase)
    las14 = unittest.TestLoader().loadTestsFromTestCase(LasV_14TestCase)
    laz = unittest.TestLoader().loadTestsFromTestCase(LazReaderTestCase)
    open_modes = unittest.TestLoader().loadTestsFromTestCase(LazNoPointMapTestCase)

    return unittest.TestSuite([reader, writer, header_writer, write_mode,
                               las13, las14, open_modes, laz])


def really_copyfile(src, dst, max_=1):
    """
        Hack for Windows when quickly creating and deleting files.
        :param src: passed to shutil.copyfile
        :param dst: passed to shutil.copyfile
        :param max_: max seconds to wait
        """
    wait = 0.01
    while not os.path.exists(dst):
        try:
            shutil.copyfile(src, dst)
        except IOError:
            time.sleep(wait)
            max_ -= wait
            if max_ <= 0:
                break


def really_remove(path, max_=1):
    """
    Hack for Windows when quickly creating and deleting files.
    os.remove can return when Windows still thinks the file exists.
    When trying to re-create the file with the same name, a PermissionError occurs.
    :param path: path to remove
    :param max_: max seconds to wait
    """
    wait = 0.01
    while os.path.exists(path):
        try:
            os.remove(path)
        except WindowsError:
            time.sleep(wait)
            max_ -= wait
            if max_ <= 0:
                break



import laspy
import numpy as np
import random
import sys

# OpenGL library doesn't support absolute import for gl, glu, glut, or arrays
import OpenGL.GL as gl 
import OpenGL.GLU as glu 
import OpenGL.GLUT as glut 
from OpenGL.arrays import vbo



# Skip debugging for speedup
#OpenGL.ERROR_CHECKING = False

# Skip logging for speedup 
#OpenGL.ERROR_LOGGING = False


def run_glviewer(file_object, mode, dim):
    glviewer = pcl_image(file_object, mode, dim)
    return(0)

class VBO_Provider():
    def __init__(self, file_object, vbsize, means, mode, dim, scaled = True):
        self.vbos = []
        self.allcolor = False
        start_idx = 0
        self.file_object = file_object
        end_idx = vbsize
        i = 1

        while(start_idx < len((file_object))):
            i += 1
            try:
                end_idx = min(len(file_object), start_idx + vbsize) 
                print("Buffering points " + str(start_idx) + " to " + str((end_idx)))
                dat = self.slice_file(start_idx, end_idx, means, scaled)
                self.set_color_mode(mode,dim, start_idx, end_idx, dat)
                _vbo = vbo.VBO(data = np.array(dat, dtype = np.float32),
                            usage = gl.GL_DYNAMIC_DRAW, target = gl.GL_ARRAY_BUFFER)
                self.vbos.append((_vbo, end_idx -start_idx))
                start_idx += vbsize
            except Exception as err:
                print("Error initializing VBO:")
                print(err)


    def slice_file(self,start_idx, end_idx, means, scaled):
        if scaled:
            return(np.array(np.vstack((self.file_object.x[start_idx:end_idx], self.file_object.y[start_idx:end_idx], self.file_object.z[start_idx:end_idx], 
                                  np.zeros(end_idx - start_idx),np.zeros(end_idx - start_idx),np.zeros(end_idx - start_idx))).T) - means)
        else:
            scale = np.array(self.file_object.header.scale + [0,0,0], dtype = np.float64)
            dat = (np.array(np.vstack((self.file_object.X[start_idx:end_idx], self.file_object.Y[start_idx:end_idx], self.file_object.Z[start_idx:end_idx], 
                                  np.zeros(end_idx - start_idx),np.zeros(end_idx - start_idx),np.zeros(end_idx - start_idx))).T) - means)
            dat *= (scale*100)
            return(dat)


    def bind(self):
        for _vbo in self.vbos:
            _vbo[0].bind()

    def unbind(self):
        for _vbo in self.vbos:
            _vbo[0].unbind()
    def draw(self):

        for _vbo in self.vbos:
            _vbo[0].bind()
            gl.glVertexPointer(3, gl.GL_FLOAT, 24,_vbo[0])
            gl.glColorPointer(3, gl.GL_FLOAT, 24, _vbo[0] + 12)
            gl.glDrawArrays(gl.GL_POINTS, 0, _vbo[1]) 
            _vbo[0].unbind()
        #gl.glMultiDrawArrays(gl.GL_POINTS, 0,100000, len(self.vbos))
    
    def set_color_mode(self, mode, dim,start_idx, end_idx, data): 
        if (mode == "default"):
            if (all([x in self.file_object.point_format.lookup for x in ("red", "green", "blue")])):
                idx_step = max(int((len(self.file_object)/1000)), 1)
                if ((all(self.file_object.red[0:len(self.file_object):idx_step] == 0)) and
                    (all(self.file_object.green[0:len(self.file_object):idx_step] == 0)) and
                    (all(self.file_object.blue[0:len(self.file_object):idx_step] == 0))):
                    print("Warning: Color data appears empty, using intensity mode. Specify -mode=rgb to override")
                    mode = "intensity"
                else:
                    mode="rgb"
            else:
                mode = "intensity" 
        if mode == "rgb" and not "red" in self.file_object.point_format.lookup:
            print("Color data not found in file, using intensity")
            mode = "intensity"
        if mode in ["grey", "greyscale", "intensity"]:
            if type(self.allcolor) == bool:
                self.allcolor = self.file_object.reader.get_dimension(dim)/float(np.max(self.file_object.reader.get_dimension(dim)))
            scaled = self.allcolor[start_idx:end_idx] + 0.1
            col = np.array((np.vstack((scaled, scaled, scaled)).T), dtype = np.float32)
            data[:,3:6] += col
            return(data)
        elif (mode == "elevation" or (mode == "heatmap" and dim == "z")):
            if type(self.allcolor) == bool:
                self.allcolor = self.heatmap(self.file_object.z)
            col = self.allcolor[start_idx:end_idx]
            data[:,3:6] += col
            return(data)
        elif (mode == "heatmap" and dim != "z"):
            if type(self.allcolor) == bool:
                self.allcolor = self.heatmap(self.file_object.reader.get_dimension(dim))
            col = self.allcolor[start_idx:end_idx]
            data[:,3:6] += col
            return(data)
        elif mode == "rgb":
            _max = max(np.max(self.file_object.red), np.max(self.file_object.green), np.max(self.file_object.blue))
            _min = min(np.min(self.file_object.red), np.min(self.file_object.green), np.min(self.file_object.blue))
            diff = _max - _min
            col = np.array(np.vstack((self.file_object.red[start_idx:end_idx], self.file_object.green[start_idx:end_idx], self.file_object.blue[start_idx:end_idx])).T, dtype = np.float32)
            col -= _min
            col /= diff
            data[:,3:6] += col
            return(data)

    def heatmap(self, vec, mode = 1):
        _max = np.max(vec)
        _min = np.min(vec)
        diff = _max-_min
        red = (vec-_min)/float(diff) 
        if mode == 1:
            col = np.array(np.vstack((red**4, np.sqrt(0.0625-(0.5-red)**4) , (1-red)**4)),dtype = np.float32).T 
        else:
            col = np.array(np.vstack((red**4, np.zeros(self.N) , (1-red)**4)),dtype = np.float32).T 
        return(col)

class pcl_image():
    def __init__(self, file_object, mode, dim):
        self.file_object = file_object
        self.read_data(mode, dim)
        self.movement_granularity = 1.0
        self.look_granularity = 16.0
        self.main()

    def main(self): 
        self.location = np.array([0.0,0.0,1500.0])
        self.focus = np.array([0.0,0.0,0.0])
        self.up = np.array([1.0,0.0,0.0])

        self.mousex = 0
        self.mousey = 0
        self.mouse_drag = gl.GL_FALSE

        # Wire up GL
        glut.glutInit(sys.argv)

        glut.glutInitDisplayMode(glut.GLUT_RGB | glut.GLUT_DOUBLE | glut.GLUT_DEPTH)
        glut.glutInitWindowSize(500,500)
        glut.glutInitWindowPosition(10,10)
        glut.glutCreateWindow("Laspy+OpenGL Pointcloud")
        glut.glutDisplayFunc(self.display)
        glut.glutReshapeFunc(self.reshape)
        glut.glutMouseFunc(self.mouse)
        glut.glutMotionFunc(self.mouse_motion)
        glut.glutKeyboardFunc(self.keyboard)
        gl.glClearColor(0.0,0.0,0.0,1.0)
        glut.glutTimerFunc(10,self.timerEvent,1)

        glut.glutMainLoop()
        return 0
 
    def read_data(self, mode, dim):
        if (np.max(self.file_object.x) - np.min(self.file_object.x)) < 1:
            means = np.array([np.mean(self.file_object.X, dtype = np.float64), 
                          np.mean(self.file_object.Y, dtype = np.float64), 
                          np.mean(self.file_object.Z, dtype = np.float64),
                          0,0,0])
            scaled = False
        else:
            means = np.array([np.mean(self.file_object.x, dtype = np.float64), 
                    np.mean(self.file_object.y, dtype = np.float64), 
                    np.mean(self.file_object.z, dtype = np.float64),
                    0,0,0])
            scaled = True
        
        self.N = len(self.file_object)
        self.data_buffer = VBO_Provider(self.file_object, 1000000, means, mode, dim, scaled) 

 

    def reshape(self, w, h):
        print("Reshape " + str(w) + ", " + str(h))
        ratio = w if h == 0 else float(w)/h
        gl.glMatrixMode(gl.GL_PROJECTION)
        gl.glLoadIdentity()
        gl.glViewport(0,0,w,h)
        gl.glLoadIdentity()
        glu.gluPerspective(90,float(ratio),0.001,3000);

        gl.glMatrixMode(gl.GL_MODELVIEW)
        
    def timerEvent(self, arg):
        # Do stuff
        glut.glutPostRedisplay()
        glut.glutTimerFunc(10,self.timerEvent,1)

    def draw_points(self, num):

        gl.glEnableClientState(gl.GL_VERTEX_ARRAY)
        gl.glEnableClientState(gl.GL_COLOR_ARRAY)
        self.data_buffer.draw()

        gl.glDisableClientState(gl.GL_COLOR_ARRAY)
        gl.glDisableClientState(gl.GL_VERTEX_ARRAY)
         
        
            
    def rotate_vector(self, vec_rot, vec_about, theta):
        d = np.sqrt(vec_about.dot(vec_about))
        
        L = np.array((0,vec_about[2], -vec_about[1], 
                    -vec_about[2], 0, vec_about[0],
                    vec_about[1], -vec_about[0], 0))
        L.shape = (3,3)

        
        try:
           R = (np.identity(3) + np.sin(theta)/d*L +
                    (1-np.cos(theta))/(d*d)*(L.dot(L)))
        except:
            print("Error in rotation.")
            return()
        return(vec_rot.dot(R))

    def display(self):
        gl.glClear(gl.GL_COLOR_BUFFER_BIT | gl.GL_DEPTH_BUFFER_BIT)
        gl.glLoadIdentity()
        glu.gluLookAt(self.location[0], self.location[1], self.location[2], 
                      self.focus[0],self.focus[1], self.focus[2] ,
                      self.up[0], self.up[1], self.up[2])
        self.draw_points(self.N)
        glut.glutSwapBuffers()

    def camera_reset(self):
        self.location = np.array([0.0,0.0,1500.0])
        self.focus = np.array([0.0,0.0,0.0])
        self.up = np.array([1.0,0.0,0.0])



    def camera_move(self,ammount, axis = 1):
        if axis == 1:
            pointing = self.focus - self.location
            pnorm = np.sqrt(pointing.dot(pointing))
            pointing /= pnorm
            self.location = self.location + ammount*pointing
            self.focus = self.location + pnorm*pointing
        elif axis == 2:
            pointing = self.focus - self.location
            direction = np.cross(self.up, pointing)
            direction /= np.sqrt(direction.dot(direction))
            self.location = self.location + ammount * direction
            self.focus = self.location + pointing
            
    def camera_yaw(self, theta):
        pointing = self.focus - self.location
        newpointing = self.rotate_vector(pointing, self.up, theta)
        self.focus = newpointing + self.location

    def camera_roll(self, theta):
        self.up = self.rotate_vector(self.up, self.focus-self.location, theta)

    def camera_pitch(self,theta):
        pointing = self.focus - self.location
        axis = np.cross(self.up, pointing)
        newpointing = self.rotate_vector(pointing, axis, theta)
        self.focus = newpointing + self.location
        self.up = np.cross(newpointing, axis)
        self.up /= np.sqrt(self.up.dot(self.up))

    def mouse(self, button, state, x, y):
        if button == glut.GLUT_LEFT_BUTTON:
            if state == glut.GLUT_DOWN:
                self.mouse_drag = gl.GL_TRUE
                self.mousex = x
                self.mousey = y
            elif state == glut.GLUT_UP and self.mouse_drag:
                self.mouse_drag = gl.GL_FALSE
        elif button == 3:
            #Scoll up
            pass
        elif button == 4:
            #Scroll down
            pass

    def mouse_motion(self,x,y):
        if self.mouse_drag:
            self.mousex = x
            self.mousey = y

    def keyboard(self,key, x, y):
        ## Looking
        if key == b"a":
            self.camera_yaw(np.pi/(self.look_granularity))
        elif key == b"d":
            self.camera_yaw(-np.pi/self.look_granularity)
        elif key == b"w":
            self.camera_pitch(-np.pi/self.look_granularity)
        elif key == b"s":
            self.camera_pitch(np.pi/self.look_granularity)
        elif key == b"e":
            self.camera_roll(np.pi/self.look_granularity)
        elif key == b"q":
            self.camera_roll(-np.pi/self.look_granularity)
        ## Moving
        elif key == b"W":
            self.camera_move(self.movement_granularity * 100.0)
        elif key == b"S":
            self.camera_move(self.movement_granularity *-100.0)
        elif key == b"A":
            self.camera_move(self.movement_granularity * 100.0, axis = 2)
        elif key == b"D":
            self.camera_move(self.movement_granularity * -100.0, axis = 2)

        elif key in (b"R", b"r"):
            self.camera_reset()


        elif key == b"+":
            self.movement_granularity *= 0.8
            self.look_granularity /= 0.8
        elif key == b"-":
            self.movement_granularity /= 0.8
            self.look_granularity *= 0.8
        elif key in (b"x", b"y", b"z"):
            self.set_up_axis(key)
        print(key)
        pass

    def set_up_axis(self, key):
        if key == b"x":
            self.up = np.array([1.0, 0.0001, 0.0001])
            self.focus[0] = self.location[0]
        elif key == b"y":
            self.up = np.array([0.0001, 1.0, 0.0001])
            self.focus[1] = self.location[1]
        elif key == b"z":
            self.up = np.array([0.0001, 0.0001, 1.0])
            self.focus[2] = self.location[2]
        if all(self.focus == self.location):
            self.focus[{"x":1, "y":2, "z":0}[key]] += 1500




import argparse
import laspy
import numpy as np
import math

def main():
    parser =argparse.ArgumentParser(description = """Open a file in rw mode and add random noise to X Y and Z.""")
    parser.add_argument("in_file", metavar = "in_file", 
                        type=str,nargs="+",help = "LAS file to screw with.")
    parser.add_argument("--x_pct", metavar = "x_pct", type = float,default = 10, help = "Percent Of X Points to jiggle")
    parser.add_argument("--y_pct", metavar = "y_pct", type = float,default = 10, help = "Percent Of X Points to jiggle")
    parser.add_argument("--z_pct", metavar = "z_pct", type = float,default = 10, help = "Percent Of X Points to jiggle")
    parser.add_argument("--x_amt", metavar = "x_amt", type = int, default = 3, help = "Ammount to jiggle selected X points.") 
    parser.add_argument("--y_amt", metavar = "y_amt", type = int, default = 3, help = "Ammount to jiggle selected Y points.") 
    parser.add_argument("--z_amt", metavar = "z_amt", type = int, default = 3, help = "Ammount to jiggle selected Z points.") 


    args = parser.parse_args()



    inFile = laspy.file.File(args.in_file[0], mode = "rw")

    def gen_noise(pct, length, amt):
        out = np.zeros(length)
        indices = np.random.random_integers(0,length-1, math.floor(float(pct)/100*length))
        vals = np.random.random_integers(-amt,amt, len(indices))
        out[indices] += vals
        return(out)

    inFile.X += gen_noise(args.x_pct, len(inFile),args.x_amt)
    inFile.Y += gen_noise(args.y_pct, len(inFile),args.y_amt)
    inFile.Z += gen_noise(args.z_pct, len(inFile),args.z_amt)
    inFile.header.update_min_max()
    inFile.close(ignore_header_changes=True)






import argparse
import laspy

class lasview():
    def __init__(self):
        self.parse_args()
        self.setup()

    def parse_args(self):
        parser =argparse.ArgumentParser(description = """Open a file in read mode and
                                        print a simple description.""")
        parser.add_argument("in_file", metavar = "in_file", 
                            type=str,nargs="+",help = "LAS file to plot")
        parser.add_argument("--mode",metavar="viewer_mode", type=str,default="default", 
                help = "Color Mode. Values to specify with a dimension: greyscale, heatmap.  Values which include a dimension: elevation, intensity, rgb")
        parser.add_argument("--dimension", metavar = "dim", type=str, default="intensity",
                help = "Color Dimension. Can be any single LAS dimension, default is intensity. Using color mode rgb, elevation, and intensity overrides this field.")

        self.args = parser.parse_args()
     
    def setup(self):
    # Check mode
        print("Reading: " + self.args.in_file[0])
        self.mode = self.args.mode
        self.dim = self.args.dimension
        try:
            inFile = laspy.file.File(self.args.in_file[0], mode = "r")
            self.inFile = inFile
        except Exception as error:
            print("Error while reading file:")
            print(error)
            quit()
        
    def view(self):
        self.inFile.visualize(self.mode, self.dim)

def main():
    expl = lasview()
    expl.view()





import argparse
import laspy
import code


class lasexplorer():
    def __init__(self):
        self.parse_args()
        self.setup()

    def parse_args(self):
        parser =argparse.ArgumentParser(description = """Open a file in read mode and
                                        print a simple description.""")
        parser.add_argument("in_file", metavar = "in_file", 
                            type=str,nargs="+",help = "LAS file to explore")
        parser.add_argument("--mode",metavar="file mode", type=str,default="r", 
                help = "Mode, default is r. Acceptable values: (r, rw)")
        parser.add_argument("-q", metavar="quiet", type=bool, default=False, 
                help ="Skip summary? (default false)")

        self.args = parser.parse_args()
     
    def setup(self):
    # Check mode
        if self.args.mode =="rw":
            print("Mode = %s, changes to file will be committed to disk." % self.args.mode)
        elif self.args.mode != "r":
            print("Warning: invalud mode: " + self.args.mode)

    # Try to read in file
        print("Reading: " + self.args.in_file[0])
        try:
            inFile = laspy.file.File(self.args.in_file[0], mode = self.args.mode)
            self.inFile = inFile
            READ_SUCCESS = True
            print("Read successful, file object is called inFile")
        except Exception as error:
            print("Error while reading file:")
            print(error)
            READ_SUCCESS = False

    # If we have a file and user wants verbose, print summary.

        if READ_SUCCESS and not self.args.q:
            print("LAS File Summary:")
            print("File Version: " + str(inFile.header.version))
            print("Point Format: " + str(inFile.header.data_format_id))
            print("Number of Point Records: " + str(len(inFile)))
            print("Point Dimensions: ")
            for dim in inFile.point_format:
                print("    " + dim.name)
        
    def explore(self):
        inFile = self.inFile
        interp = code.InteractiveConsole(locals={"inFile":inFile}) 
        try:
            interp.interact()
        except (KeyboardInterrupt, SystemExit):
            quit()

def main():
    expl = lasexplorer()
    expl.explore()





import mmap
import laspy
import os
import datetime
import struct
from types import GeneratorType
import numpy as np
import copy
from distutils.version import StrictVersion

try:
    import lazperf
    HAVE_LAZPERF = True
except ImportError:
    HAVE_LAZPERF = False

# Not used right now - but could be a handy place to centralize file modes
FILE_MODES = ["r-", "r", "rw", "w"]

try:
    xrange
except NameError:
    xrange = range

try:
    buffer
except NameError:
    buffer = memoryview

def _prepare_np_frombuffer_data(data):
    # convert data according to `numpy` changes:
    # https://github.com/numpy/numpy/blob/a9bb517554004cf2ce7a4be93bcbfb63ee149844/doc/source/release/1.17.0-notes.rst#do-not-lookup-__buffer__-attribute-in-numpyfrombuffer
    change_data = isinstance(data, FakeMmap) and \
        StrictVersion(np.version.version) >= StrictVersion('1.17.0')
    return data.view if change_data else data


def read_compressed(filename):
    import subprocess

    laszip_names = ('laszip', 'laszip.exe', 'laszip-cli', 'laszip-cli.exe')
    laszip_binary = ''

    for binary in laszip_names:
        in_path = [os.path.isfile(os.path.join(x, binary))for x in os.environ["PATH"].split(os.pathsep)]
        if any(in_path):
            laszip_binary = binary
            break
    else:
        raise(laspy.util.LaspyException("laszip executable (%s) was not found in PATH" % ", ".join(laszip_names)))

    prc=subprocess.Popen([laszip_binary, "-olas", "-stdout", "-i", filename],
                stdout=subprocess.PIPE, stderr=subprocess.PIPE, bufsize=-1)
    data, stderr=prc.communicate()
    if prc.returncode != 0:
        # What about using the logging module instead of prints?
        print("Unusual return code from %s: %d" % (laszip_binary, prc.returncode))
        if stderr and len(stderr)<2048:
            print(stderr)
        raise ValueError("Unable to read compressed file!")
    return data


class FakeMmap(object):
    '''
    An object imitating a memory mapped file,
    constructed from 'buffer like' data.
    '''
    def __init__(self, filename, pos=0):
        data = read_compressed(filename)
        self.view = memoryview(data)
        self.pos = pos
        # numpy needs this, unfortunately
        # Note: this is a memoryview in python3. Does numpy still "need" this?
        self.__buffer__ = buffer(data)

    def __len__(self):
        return len(self.view)

    def __getitem__(self, i):
        return self.view[i]

    def close(self):
        self.view = None

    def flush(self):
        pass

    def seek(self, nbytes, whence=0):
        if whence == 0:
            self.pos = nbytes
        else:
            self.pos += nbytes

    def read(self, nbytes):
        try:
            nbytes = int(nbytes)
        except ValueError:
            nbytes = 0
        out = self.view[self.pos:self.pos+nbytes]
        self.pos += nbytes
        return(out)

    def tell(self):
        return self.pos

    def size(self):
        return len(self.view)


class DataProvider():
    '''Provides access to the file object, the memory map, and the numpy point map.'''
    def __init__(self, filename, manager):
        '''Construct the data provider. _mmap refers to the memory map, and _pmap
        refers to the numpy point map.'''
        self.filename = filename
        self.fileref = False
        self._mmap = False
        self._pmap = False
        self._evlrmap = False
        self.manager = manager
        self.mode = manager.mode
        # Figure out if this file is compressed
        if self.mode in ("w"):
            self.compressed = False
        else:
            try:
                tmpref = open(filename, "rb")
                tmpref.seek(104)
                fmt = int(struct.unpack("<B", tmpref.read(1))[0])
                compression_bit_7 = (fmt & 0x80) >> 7
                compression_bit_6 = (fmt & 0x40) >> 6
                if (not compression_bit_6 and compression_bit_7):
                    self.compressed = True
                else:
                    self.compressed = False
                tmpref.close()
            except Exception as e:
                raise laspy.util.LaspyException("Error determining compression: "
                        + str(e))

    def open(self, mode):
        '''Open the file, catch simple problems.'''
        self.fileref = open(self.filename, mode)

    def get_point_map(self, informat):
        '''Get point map is used to build and return a numpy frombuffer view of the mmapped data,
        using a valid laspy.util.Format instance for the desired point format. This method is used
        to provide access to extra_bytes even when dimensions have been explicitly defined via an
        extra_bytes VLR record.'''
        if type(self._mmap) == bool:
            self.map()
        self.pointfmt = np.dtype([("point", zip([x.name for x in informat.specs],
                                [x.np_fmt for x in informat.specs]))])
        if not self.manager.header.version in ("1.3", "1.4"):
            np_frombuffer_data = _prepare_np_frombuffer_data(self._mmap)
            _pmap = np.frombuffer(np_frombuffer_data, self.pointfmt,
                        offset = self.manager.header.data_offset)
        else:
            np_frombuffer_data = _prepare_np_frombuffer_data(self._mmap)
            _pmap = np.frombuffer(np_frombuffer_data, self.pointfmt,
                        offset = self.manager.header.data_offset,
                        count = self.manager.header.point_records_count)
        return(_pmap)

    def point_map(self):
        '''Create the numpy point map based on the point format.'''
        if type(self._mmap) == bool:
            self.map()

        if self.mode == "r-":
            # Do not construct the point map in case
            return
        self.pointfmt = np.dtype([("point", [(str(x.name), x.np_fmt) for x in self.manager.point_format.specs])])

        if self.manager.header.version not in ("1.3", "1.4"):
            if self.compressed and HAVE_LAZPERF:
                import lazperf
                from lazperf import VLRDecompressor
                vlr = self.findLASzipVLR(self.manager.header.vlrs)
                if not vlr:
                    raise laspy.util.LaspyException("""Unable to find LASzip VLR!""")

                vlr_data = np.frombuffer(vlr.VLR_body, 
                                         np.uint8, 
                                         count = vlr.rec_len_after_header) 

                # first 8 bytes after VLRs are laszip, then is start of data
                laszip_offset = self.manager.header.data_offset + 8

                np_frombuffer_data = _prepare_np_frombuffer_data(self._mmap)
                points_compressed = np.frombuffer(np_frombuffer_data, 
                                                  np.uint8, 
                                                  offset = laszip_offset, 
                                                  count = self._mmap.size() - laszip_offset)

                record_length = self.manager.header.data_record_length
                decompressor = lazperf.VLRDecompressor( points_compressed, 
                                                        record_length,
                                                        vlr_data)

                uncompressed = decompressor.decompress_points(self.manager.header.point_records_count)

                # we've decompressed the points, now stick the header on the 
                # front and make a new mmap
                np_frombuffer_data = _prepare_np_frombuffer_data(self._mmap)
                header = np.frombuffer(np_frombuffer_data, np.uint8, count = self.manager.header.data_offset)
                full = np.zeros(len(header) + len(uncompressed), dtype=np.uint8)
                full[0:len(header)] = header
                full[len(header):len(uncompressed)+len(header)] = uncompressed

                self._mmap = full


            np_frombuffer_data = _prepare_np_frombuffer_data(self._mmap)
            self._pmap = np.frombuffer(np_frombuffer_data, self.pointfmt,
                                       offset=self.manager.header.data_offset)
            if self.manager.header.point_records_count != len(self._pmap):
                if self.manager.mode == "r":
                    raise laspy.util.LaspyException("""Invalid Point Records Count Information Encountered in Header.
                                        Please correct. Header.point_records_count = %i, and %i records actually detected."""%(self.manager.header.point_records_count, len(self._pmap)))
                else:
                    print("""WARNING: laspy found invalid data in header.point_records_count.
                            Header.point_records_count = %i, and %i records actually detected.
                            Attempting to correct mismatch.""") % (self.manager.header.point_records_count, len(self._pmap))
                    self.manager.header.point_records_count = len(self._pmap)
        else:
            np_frombuffer_data = _prepare_np_frombuffer_data(self._mmap)
            self._pmap = np.frombuffer(np_frombuffer_data, self.pointfmt,
                                       offset=self.manager.header.data_offset,
                                       count=self.manager.header.point_records_count)

    def close(self, flush=True):
        '''Close the data provider and flush changes if _mmap and _pmap exist.'''
        if flush and self.manager.has_point_records:
            try:
                self._mmap.flush()
                self._mmap.close()
            except(Exception):
                self._mmap = False
                self._pmap = False
        self._mmap = False
        self._pmap = False
        if self.fileref != False:
            try:
                self.fileref.close()
            except(Exception):
                raise laspy.util.LaspyException("Error closing file.")

    def findLASzipVLR(self, vlrs):
        for vlr in vlrs:
            if vlr.user_id.rstrip(' \t\r\n\0') == 'laszip encoded':
                if vlr.record_id == 22204:
                    return vlr
        return None


    def map(self):
        '''Memory map the file'''
        if self.fileref == False and not self.compressed:
            raise laspy.util.LaspyException("File not opened.")
        if self.mode in ("r", "r-"):

            if self.compressed and self.mode != "r-" and not HAVE_LAZPERF:
                 self._mmap = FakeMmap(self.filename)
            else:
                self._mmap = mmap.mmap(self.fileref.fileno(), 0, access = mmap.ACCESS_READ)
        elif self.mode in ("w", "rw"):
            self._mmap = mmap.mmap(self.fileref.fileno(), 0, access = mmap.ACCESS_WRITE)
        else:
            raise laspy.util.LaspyException("Invalid Mode: " + str(self.mode))

    def remap(self,flush = True, point_map = False):
        '''Re-map the file. Flush changes, close, open, and map. Optionally point map.'''
        if flush and type(self._mmap) != bool:
            self._mmap.flush()
        self.close(flush=False)
        if self.mode in ("r", "r-"):
            self.open("rb")
        elif self.mode in ("w", "rw"):
            self.open("r+b")
        else:
            raise laspy.util.LaspyException("Invalid Mode: " + str(self.mode))
        self.map()
        if point_map:
            self.point_map()

    def __getitem__(self, index):
        '''Return the raw bytes corresponding to the point @ index.'''
        try:
            index.stop
        except AttributeError:
            return(self._pmap[index][0])
        if index.step:
            step = index.step
        else:
            step = 1
        return([x[0] for x in self._pmap[index.start:index.stop,step]])

    def __setitem__(self, key, value):
        '''Assign raw bytes for point @ key'''
        try:
            key.stop
        except AttributeError:
            self._pmap[key] = (value,)
            return
        self._pmap[key.start:key.stop] = [(x,) for x in value]

    def filesize(self):
        '''Return the size ofs the current map'''
        if self._mmap == False:
            raise laspy.util.LaspyException("File not mapped")
        return(self._mmap.size())

class FileManager(object):
    '''Superclass of Reader and Writer, provides most of the data manipulation functionality in laspy.'''
    def __init__(self,filename, mode, header=False, vlrs=False, evlrs=False):
        '''Build the FileManager object. This is done when opening the file
        as well as upon completion of file modification actions like changing the
        header padding.'''
        self.compressed = False
        self.vlr_formats = laspy.util.Format("VLR")
        self.evlr_formats = laspy.util.Format("EVLR")
        self.mode = mode
        self.data_provider = DataProvider(filename, self)
        self.setup_memoizing()
        self.calc_point_recs = False
        self.point_refs = False
        self._current = 0
        self.padded = False
        if self.mode in ("r", "r-"):
            self.setup_read_write(vlrs,evlrs, read_only=True)
            return
        elif self.mode == "rw":
            self.setup_read_write(vlrs, evlrs, read_only=False)
            return
        elif self.mode == "w":
            self.setup_write(header, vlrs, evlrs)
            return
        else:
            raise laspy.util.LaspyException("Mode %s not supported / implemented" % mode)

    def close(self):
        '''Help the garbage collector by deleting some of the circular references'''
        self.data_provider.close()
        self.data_provider = None
        self.header = None
        self._header = None
        self.vlrs = None
        self.evlrs = None

    def setup_read_write(self, vlrs, evlrs, read_only=True):
        # Check if read only mode, if not open for updating.
        if read_only:
            open_mode = "rb"
        else:
            open_mode = "r+b"
        self._header_current = True
        self.data_provider.open(open_mode)
        self.data_provider.map()
        self.header_format = laspy.util.Format("h" + self.grab_file_version())
        self.get_header(self.grab_file_version())
        self.populate_vlrs()
        self.point_refs = False
        self.has_point_records = True
        self._current = 0
        self.correct_rec_len()
        self.compressed = self.point_format.compressed

        #if self.point_format.compressed:
        #    self.compressed = True
        #    self.data_provider.remap()
        #else:
        #    self.compressed = False
        if self.mode == "r-":
            self.extra_dimensions = []
            self.vlrs = []
            self.evlrs = []
            return
        self.data_provider.point_map()
        if self.header.version in ("1.3", "1.4"):
            #gives key error if called with buffer hack for some reason...
            self.populate_evlrs()
        else:
            self.evlrs = []
        if vlrs != False:
            self.set_vlrs(vlrs)
        if evlrs != False:
            self.set_evlrs(vlrs)

        # If extra-bytes descriptions exist in VLRs, use them.
        eb_vlrs = [x for x in self.vlrs if x.type == 1]
        eb_vlrs.extend([x for x in self.evlrs if x.type == 1])
        self.extra_dimensions = []
        if len(eb_vlrs) > 1:
            raise laspy.util.LaspyException("Only one ExtraBytes VLR currently allowed.")
        elif len(eb_vlrs) == 1:
            self.naive_point_format = self.point_format
            self.extra_dimensions = eb_vlrs[0].extra_dimensions
            new_pt_fmt = laspy.util.Format(self.point_format.fmt, extradims
                    = self.extra_dimensions)
            self.point_format = new_pt_fmt
            self.data_provider.remap(point_map = True)
        return

    def setup_write(self,header, vlrs, evlrs):
        self._header_current = False
        if header == False:
            raise laspy.util.LaspyException("Write mode requires a valid header object.")
        ## No file to store data yet.
        self.has_point_records = False
        self.data_provider.open("w+b")
        self.header_format = header.format
        self._header = header
        self.header = laspy.header.HeaderManager(header = header, reader = self)
        self.initialize_file_padding(vlrs)

        ## We have a file to store data now.
        self.data_provider.remap()
        self.header.flush()

        self.correct_rec_len()
        if not vlrs in [[], False]:
            self.set_vlrs(vlrs)
        else:
            self.vlrs = []
        if not evlrs in [[], False]:
            self.set_evlrs(evlrs)
        else:
            self.evlrs = []
        self.verify_num_vlrs()
        if self._header.created_year == 0:
            self.header.date = datetime.datetime.now()
        self.populate_vlrs()
        self.populate_evlrs()
        # If extra-bytes descriptions exist in VLRs, use them.
        eb_vlrs = [x for x in self.vlrs if x.type == 1]
        eb_vlrs.extend([x for x in self.evlrs if x.type == 1])
        self.extra_dimensions = []
        if len(eb_vlrs) > 1:
            raise laspy.util.LaspyException("Only one ExtraBytes VLR currently allowed.")
        elif len(eb_vlrs) == 1:
            self.naive_point_format = self.point_format
            self.extra_dimensions = eb_vlrs[0].extra_dimensions
            new_pt_fmt = laspy.util.Format(self.point_format.fmt, extradims
                    = self.extra_dimensions)
            self.point_format = new_pt_fmt
        return

    def verify_num_vlrs(self):
        headervlrs = self.get_header_property("num_variable_len_recs")
        calc_headervlrs = len(self.vlrs)
        if headervlrs != calc_headervlrs:
            raise laspy.util.LaspyException('''Number of EVLRs provided does not match the number
                                 specified in the header. (copied headers do not maintain
                                 references to their EVLRs, that might be your problem.
                                 You can pass them explicitly to the File constructor.)''')

        if self.header.version == "1.4":
            calc_headerevlrs = len(self.evlrs)
            headerevlrs = self.get_header_property("num_evlrs")
            if headerevlrs != calc_headerevlrs:
                raise laspy.util.LaspyException('''Number of EVLRs provided does not match the number
                                     specified in the header. (copied headers do not maintain
                                     references to their EVLRs, that might be your problem.
                                     You can pass them explicitly to the File constructor.)''')
    def correct_rec_len(self):
        extrabytes = self.header.data_record_length-laspy.util.Format(self.header.data_format_id).rec_len
        if extrabytes >= 0:
            self.point_format = laspy.util.Format(self.header.data_format_id,extra_bytes= extrabytes)
        else:
            self.point_format = laspy.util.Format(self.header.data_format_id)
            self.set_header_property("data_record_length", self.point_format.rec_len)

    def initialize_file_padding(self, vlrs):
        filesize = self._header.format.rec_len
        self._header.header_size = filesize
        if vlrs != False:
            filesize += sum([len(x) for x in vlrs])
        self.vlr_stop = filesize
        if self._header.data_offset != 0:
            filesize = max(self._header.data_offset, filesize)
        self._header.data_offset = filesize
        self.data_provider.fileref.write(b"\x00"*filesize)
        return

    def setup_memoizing(self):
        self.header_changes = set()
        self.header_properties = {}

    def populate_c_packers(self):
        '''This is depricated if the numpy point map is used, because nparr.tostring() is MUCH faster.
        This creates compiled Struct objects for various formats.
        '''
        for spec in self.point_format.specs:
            self.c_packers[spec.name] = struct.Struct(spec.fmt)
            self.c_packers[spec.fmt] = self.c_packers[spec.name]

    def packed_str(self, string):
        '''Take a little endian binary string, and convert it to a python int.'''
        return(sum([int(string[idx])*(2**idx) for idx in xrange(len(string))]))

    def binary_str(self, N, zerolen = 8):
        '''Take a python integer and create a binary string padded to len zerolen.'''
        raw_bin = bin(N)[2:][::-1]
        padding = zerolen-len(raw_bin)
        if padding < 0:
            raise laspy.util.LaspyException("Invalid Data: Packed Length is Greater than allowed.")
        return(raw_bin + '0'*(zerolen-len(raw_bin)))

    def bit_transform(self, x, low, high):
        return np.right_shift(np.bitwise_and(x, 2**high - 1), low)

    def read(self, bytes):
        '''Wrapper for mmap.mmap read function'''
        return(self.data_provider._mmap.read(bytes))

    def reset(self):
        '''Refresh the mmap and fileref'''
        self.data_provier.remap()
        return

    def seek(self, bytes, rel = True):
        '''Wrapper for mmap.mmap seek functions, make option rel explicit'''
        self._current = None
        if rel:
            self.data_provider._mmap.seek(bytes,1)
            return
        self.data_provider._mmap.seek(bytes, 0)

    def read_words(self, name, rec_type = "vlr"):
        '''Read a consecutive sequence of packed binary data, return a single
        element or list.'''
        if rec_type == "vlr":
            source = self.vlr_formats
        elif rec_type == "evlr":
            source = self.evlr_formats
        elif rec_type == "header":
            source = self.header_format
        else:
            raise laspy.util.LaspyException("Invalid source: " + str(rec_type))
        try:
            dim = source.lookup[name]
        except KeyError:
            raise laspy.util.LaspyException("Dimension " + name + " not found.")
        return(self._read_words(dim.fmt, dim.num, dim.length))

    def _read_words(self, fmt, num, bytes):
        '''Read a consecutive sequence of packed binary data, return a single
        element or list'''
        outData = []
        for i in xrange(num):
            dat = self.read(bytes)
            try:
                unpacked = struct.unpack(fmt, dat)[0]
                if fmt == '<s':
                    struct.unpack(fmt, dat)[0]
                    unpacked = unpacked.decode('ascii')
            except (struct.error, UnicodeDecodeError):
                # this is often NULs and random data that occurs after the
                # ending NUL.
                unpacked = '\x00'
            outData.append(unpacked)
        if len(outData) > 1:
            return(outData)
        return(outData[0])

    def _pack_words(self, fmt, num, bytes, val):
        if num == 1:
            return(struct.pack(fmt, val))
        outData = "".join([struct.pack(fmt, val[i]) for i in xrange(num)])
        return(outData)


    def grab_file_version(self):
        '''Manually grab file version from header'''
        self.seek(24, rel = False)
        v1 = self._read_words("<B", 1, 1)
        v2 = self._read_words("<B", 1, 1)
        self.seek(0, rel = False)
        return(str(v1) +"." +  str(v2))

    def get_header(self, file_version = 1.2):
        '''Return the header object, or create one if absent.'''
        ## Why is this != neccesary?
        try:
            return(self.header)
        except:
            self.header = laspy.header.HeaderManager(header = laspy.header.Header(file_version), reader = self)
            return(self.header)

    def populate_evlrs(self):
        '''Catalogue the extended variable length records'''
        self.evlrs = []

        if not self.header.version in ("1.3", "1.4"):
            return


        if self.header.version == "1.3":
            if self.header.start_wavefm_data_rec != 0:
                self.seek(self.header.start_wavefm_data_rec, rel = False)
                num_vlrs = 1
            else:
                num_vlrs = 0
        elif self.header.version == "1.4":
            self.seek(self.header.start_first_evlr, rel = False)
            num_vlrs = self.get_header_property("num_evlrs")
        for i in xrange(num_vlrs):
            new_vlr = laspy.header.EVLR(None, None, None)
            new_vlr.build_from_reader(self)
            self.evlrs.append(new_vlr)
        return


    def populate_vlrs(self):
        '''Catalogue the variable length records'''
        self.vlrs = []
        self.seek(self.header.header_size, rel = False)
        for i in xrange(self.get_header_property("num_variable_len_recs")):
            new_vlr = laspy.header.VLR(None, None, None)
            new_vlr.build_from_reader(self)
            self.vlrs.append(new_vlr)
            if self.data_provider._mmap.tell() > self.header.data_offset:
                self.seek(self.header.data_offset, rel = False)
                raise laspy.util.LaspyException("Error, Calculated Header Data "
                    "Overlaps The Point Records!")
        self.vlr_stop = self.data_provider._mmap.tell()
        return

    def get_vlrs(self):
        '''Populate and return list of :obj:`laspy.header.VLR` objects`.'''
        try:
            return(self.vlrs)
        except:
            self.populate_vlrs()
            return(self.vlrs)

    def push_vlrs(self):
        self.set_vlrs(self.vlrs)

    def get_evlrs(self):
        try:
            return(self.evlrs)
        except:
            self.populate_evlrs()
            return(self.evlrs)

    def get_padding(self):
        '''Return the padding between the end of the VLRs and the beginning of
        the point records'''
        return(self.header.data_offset - self.vlr_stop)

    def get_pointrecordscount(self):
        '''calculate the number of point records'''
        return(self.get_header_property("point_records_count"))

    def set_input_srs(self):
        pass

    def set_output_srsS(self):
        pass

    def get_raw_point_index(self,index):
        '''Return the byte index of point number index'''
        return(self.header.data_offset +
            index*self.header.data_record_length)

    def get_points(self):
        '''Return a numpy array of all point data in a file.'''
        if not self.has_point_records:
            return None
        # We don't need to build point_refs here (perhaps not anymore, at all)
        # if type(self.point_refs) == bool:
        #    self.build_point_refs()
        #single_fmt = self.point_format.pt_fmt_long[1:]
        #fmtlen = len(single_fmt)
        #big_fmt_string = "".join(["<", single_fmt*self.header.point_records_count])
        #pts =  unpack(big_fmt_string, self.data_provider._mmap[self.header.data_offset:self.data_provider._mmap.size()])
        #return((laspy.util.Point(self, unpacked_list = pts[fmtlen*i:fmtlen*(i+1)]) for i in xrange(self.header.point_records_count)))
        #return([laspy.util.Point(self,x) for x in self._get_raw_dimension(0, self.header.data_record_length)])
        #return((x[0] for x in self.data_provider._pmap))
        return(self.data_provider._pmap)

    def get_raw_point(self, index):
        '''Return the raw bytestring associated with point of number index'''
        #start = (self.header.data_offset +
        #    index * self.header.data_record_length)
        #return(self.data_provider._mmap[start : start +
        #     self.header.data_record_length])
        return(self.data_provider._pmap[index][0].tostring())


#self, reader, startIdx ,version
    def get_point(self, index, nice=False):
        '''Return point object for point of number index / #legacy_api'''
        if index >= self.get_pointrecordscount():
            return
        self._current = index
        return(laspy.util.Point(self, self.get_raw_point(index), nice= nice))


    def get_next_point(self):
        '''Return next point object via get_point / #legacy_api'''
        if self._current == None:
            raise laspy.util.LaspyException("No Current Point Specified," +
                            " use Reader.GetPoint(0) first")
        if self._current == self.get_pointrecordscount():
            return
        return self.get_point(self._current + 1)

    def build_point_refs(self):
        '''Build array of point offsets '''
        # Is this really needed?
        pts = int(self.get_pointrecordscount())
        length = int(self.header.data_record_length)
        offs = int(self.header.data_offset)
        self.point_refs = [x*length + offs
                for x in xrange(pts)]
        return

    def get_dimension(self, name):
        '''Grab a point dimension by name, returning a numpy array. Refers to
        reader.point_format for the required Spec instance.'''
        if not self.has_point_records:
            return None
        #if type(self.point_refs) == bool:
        #    self.build_point_refs()
        if type(self.data_provider._pmap) == bool:
            self.data_provider.point_map()
        try:
            spec = self.point_format.lookup[name]
            #return(self._get_dimension(spec))
            return(self._get_dimension(spec))
        except KeyError:
            raise laspy.util.LaspyException("Dimension: " + str(name) +
                            " not found. Existing dimensions: " + ", ".join(self.point_format.lookup.keys()))

    def _get_dimension(self, spec):
        return(self.data_provider._pmap["point"][spec.name])

    def _get_dimension_by_specs(self,offs, fmt, length):
        '''Return point dimension of specified offset format and length'''
        if type(self.point_refs) == bool:
            self.build_point_refs()
        _mmap = self.data_provider._mmap
        prefs = (offs + x for x in self.point_refs)
        packer = self.c_packers[fmt]
        return((packer.unpack(_mmap[x:x+length])[0] for x in prefs))




    def _get_raw_dimension(self,spec):
        '''Return point dimension of specified offset format and length'''
        #_mmap = self.data_provider._mmap
        #prefs = (offs + x for x in self.point_refs)
        #return((_mmap[start + offs : start+offs+length] for start in prefs))
        return(self.data_provider._pmap["point"][spec.name].tostring())

    def _get_raw_datum(self, rec_offs, spec):
        '''return raw bytes associated with non dimension field (VLR/Header)'''
        return(self.data_provider._mmap[(rec_offs + spec.offs):(rec_offs + spec.offs
                        + spec.num*spec.length)])

    def _get_datum(self, rec_offs, spec):
        '''Return unpacked data assocaited with non dimension field (VLR/Header)'''
        data = self._get_raw_datum(rec_offs, spec)
        if spec.num == 1:
            return(struct.unpack(spec.fmt, data)[0])
        unpacked = map(lambda x: struct.unpack(spec.fmt,
            data[x*spec.length:(x+1)*spec.length])[0], xrange(spec.num))
        if spec.pack:
            # return("".join([str(x[0]) for x in unpacked]))
            if spec.fmt == '<s':
                s = ''.join(x.decode('ascii') for x in unpacked)
                return s
            else:
                return("".join([str(x[0]) for x in unpacked]))
        else:
            unpacked = list(unpacked) # evaluate the map in python3
        return(unpacked)

    def get_raw_header_property(self, name):
        '''Wrapper for grabbing raw header bytes with _get_raw_datum'''
        spec = self.header_format.lookup[name]

        return(self._get_raw_datum(0, spec))

    def get_header_property(self, name):
        '''Wrapper for grabbing unpacked header data with _get_datum'''
        if name in self.header_changes:
            spec = self.header_format.lookup[name]
            new_val = self._get_datum(0, spec)
            self.header_properties[name] = new_val
            self.header_changes.remove(name)
            return(new_val)
        elif name in self.header_properties:
            return(self.header_properties[name])
        else:
            spec = self.header_format.lookup[name]
            val = self._get_datum(0, spec)
            self.header_properties[name] = val
            return(val)

    def get_x(self, scale=False):
        if not scale:
            return(self.get_dimension("X"))
        return(self.get_dimension("X")*self.header.scale[0] + self.header.offset[0])

    def get_y(self, scale=False):
        if not scale:
            return(self.get_dimension("Y"))
        return(self.get_dimension("Y")*self.header.scale[1] + self.header.offset[1])

    def get_z(self, scale=False):
        if not scale:
            return(self.get_dimension("Z"))
        return(self.get_dimension("Z")*self.header.scale[2] + self.header.offset[2])

    def get_intensity(self):
        return(self.get_dimension("intensity"))

    def get_flag_byte(self):
        return(self.get_dimension("flag_byte"))

    def get_raw_classification_flags(self):
        return(self.get_dimension("classification_flags"))

    def get_classification_flags(self):
        if not self.header.data_format_id in (6,7,8,9,10):
            return(self.get_classification())
        rawDim = self.get_raw_classification_flags()
        return self.bit_transform(rawDim, 0, 4)

    def get_classification_byte(self):
        return(self.get_dimension("classification_byte"))

    def get_return_num(self):
        rawDim = self.get_flag_byte()
        if self.header.data_format_id in (0,1,2,3,4,5):
            return self.bit_transform(rawDim, 0, 3)
        elif self.header.data_format_id in (6,7,8,9,10):
            return self.bit_transform(rawDim, 0, 4)

    def get_num_returns(self):
        rawDim = self.get_flag_byte()
        if self.header.data_format_id in (0,1,2,3,4,5):
            return self.bit_transform(rawDim, 3, 6)
        elif self.header.data_format_id in (6,7,8,9,10):
            return self.bit_transform(rawDim, 4, 8)

    def get_scanner_channel(self):
        rawDim = self.get_raw_classification_flags()
        if not self.header.data_format_id in (6,7,8,9,10):
            raise laspy.util.LaspyException("Scanner Channel not present for point format: " + str(self.header.data_format_id))
        return self.bit_transform(rawDim, 4, 6)

    def get_scan_dir_flag(self):
        if self.header.data_format_id in (0,1,2,3,4,5):
            rawDim = self.get_flag_byte()
        elif self.header.data_format_id in (6,7,8,9,10):
            rawDim = self.get_raw_classification_flags()
        return self.bit_transform(rawDim, 6, 7)

    def get_edge_flight_line(self):
        if self.header.data_format_id in (0,1,2,3,4,5):
            rawDim = self.get_flag_byte()
        elif self.header.data_format_id in (6,7,8,9,10):
            rawDim = self.get_raw_classification_flags()
        return self.bit_transform(rawDim, 7, 8)

    def get_raw_classification(self):
        return(self.get_dimension("raw_classification"))

    def get_classification(self):
        if self.header.data_format_id in (0,1,2,3,4,5):
            return self.bit_transform(self.get_raw_classification(), 0, 5)
        elif self.header.data_format_id in (6,7,8,9,10):
            return(self.get_dimension("classification_byte"))

    def get_synthetic(self):
        if self.header.data_format_id in (6,7,8,9,10):
            return self.bit_transform(self.get_raw_classification_flags(), 0, 1)

        return self.bit_transform(self.get_raw_classification(), 5, 6)

    def get_key_point(self):
        if self.header.data_format_id in (6,7,8,9,10):
            return self.bit_transform(self.get_raw_classification_flags(), 1, 2)

        return self.bit_transform(self.get_raw_classification(), 6, 7)

    def get_withheld(self):
        if self.header.data_format_id in (6,7,8,9,10):
            return self.bit_transform(self.get_raw_classification_flags(), 2, 3)

        return self.bit_transform(self.get_raw_classification(), 7, 8)

    def get_overlap(self):
        if self.header.data_format_id in (6,7,8,9,10):
            return self.bit_transform(self.get_raw_classification_flags(), 3, 4)
        else:
            raise laspy.util.LaspyException("Overlap only present in point formats > 5.")

    def get_scan_angle_rank(self):
        return(self.get_dimension("scan_angle_rank"))

    def get_scan_angle(self):
        return(self.get_dimension("scan_angle"))

    def get_user_data(self):
        return(self.get_dimension("user_data"))

    def get_pt_src_id(self):
        return(self.get_dimension("pt_src_id"))

    def get_gps_time(self):
        return(self.get_dimension("gps_time"))

    def get_red(self):
        return(self.get_dimension("red"))

    def get_green(self):
        return(self.get_dimension("green"))

    def get_blue(self):
        return(self.get_dimension("blue"))


    def get_nir(self):
        return(self.get_dimension("nir"))


    def get_wave_packet_desc_index(self):
        return(self.get_dimension("wave_packet_desc_index"))

    def get_byte_offset_to_waveform_data(self):
        return(self.get_dimension("byte_offset_to_waveform_data"))

    def get_waveform_packet_size(self):
        return(self.get_dimension("waveform_packet_size"))

    def get_return_point_waveform_loc(self):
        return(self.get_dimension("return_point_waveform_loc"))

    def get_x_t(self):
        return(self.get_dimension("x_t"))

    def get_y_t(self):
        return(self.get_dimension("y_t"))

    def get_z_t(self):
        return(self.get_dimension("z_t"))

    def get_extra_bytes(self):
        if "extra_bytes" in self.point_format.lookup.keys():
            return(self.get_dimension("extra_bytes"))
        elif self.extra_dimensions != []:
            newmap = self.data_provider.get_point_map(self.naive_point_format)
            return(newmap["point"]["extra_bytes"])
        else:
            raise laspy.util.LaspyException("Extra bytes not present in record")


class Reader(FileManager):
    '''Just a subclass of FileManager'''
    pass

class Writer(FileManager):

    def close(self, ignore_header_changes = False, minmax_mode = "scaled"):
        '''Flush changes to mmap and close mmap and fileref'''
        if (not ignore_header_changes) and (self.has_point_records):
            if not self._header_current:
                self.header.update_histogram()
            self.header.update_min_max(minmax_mode)
        super(Writer, self).close()

    def set_evlrs(self, value):
        if value == False or len(value) == 0:
            return
        if not all([x.isEVLR for x in value]):
            raise laspy.util.LaspyException("set_evlrs requers an iterable object " +
                        "composed of :obj:`laspy.header.EVLR` objects.")
        elif self.mode == "w+":
            raise NotImplementedError
        elif self.mode in ("rw", "w"):
            if self.header.version == "1.3":
                old_offset = self.header.start_wavefm_data_rec
            elif self.header.version == "1.4":
                old_offset = self.header.start_first_evlr
                self.set_header_property("num_evlrs", len(value))
            else:
                raise laspy.util.LaspyException("Invalid File Version for EVLRs: " + str(self.header.version))
            # Good we know where the EVLRs should go... but what about if we don't have point records yet?
            # We can't make that decision yet, in case the user wants to subset the data.
            if not self.has_point_records:
                old_offset = self.header.data_offset
                if self.header.version == "1.3":
                    self.header.start_wavefm_data_rec = old_offset
                else:
                    if len(value) == 1:
                        self.header.start_first_evlr = old_offset
                        self.header.start_wavefm_data_rec = old_offset
                    else:
                        wf = self.header.start_wavefm_data_rec
                        fe = self.header.start_first_evlr
                        new_wvfm = wf - min(wf, fe) + old_offset
                        new_frst = fe - min(wf, fe) + old_offset
                        self.header.start_wavefm_data_rec = new_wvfm
                        self.header.start_first_evlr = new_frst
                #if old_offset != 0:
                #    self.pad_file_for_point_recs(self.get_pointrecordscount())
                #else:
                #    old_offset = self.header.data_offset
                #    self.pad_file_for_point_recs(self.get_pointrecordscount())

            self.data_provider.fileref.seek(0, 0)
            dat_part_1 = self.data_provider.fileref.read(old_offset)
            # Manually Close:
            self.data_provider.close(flush=False)
            self.data_provider.open("w+b")
            self.data_provider.fileref.write(dat_part_1)
            total_evlrs = sum([len(x) for x in value])
            self.data_provider.fileref.write(b"\x00"*total_evlrs)
            self.data_provider.fileref.close()
            self.data_provider.open("r+b")
            self.data_provider.map()
            self.seek(old_offset, rel = False)

            for evlr in value:
                self.data_provider._mmap.write(evlr.to_byte_string())

            if self.has_point_records:
                self.data_provider.point_map()
            self.populate_evlrs()

        else:
            raise(laspy.util.LaspyException("set_evlrs requires the file to be opened in a " +
                "write mode, and must be performed before point information is provided." +
                "Try closing the file and opening it in rw mode. "))

    def save_vlrs(self):
        self.set_vlrs(self.vlrs)

    def set_vlrs(self, value):
        if value == False or len(value) == 0:
            return
        if not all([x.isVLR for x in value]):
            raise laspy.util.LaspyException("set_vlrs requers an iterable object " +
                        "composed of :obj:`laspy.header.VLR` objects.")
        elif self.mode == "w+":
            raise NotImplementedError
        elif self.mode == "rw":
            current_size = self.data_provider._mmap.size()
            current_padding = self.get_padding()
            old_offset = self.header.data_offset
            new_offset = current_padding + self.header.header_size + sum([len(x) for x in value])
            self.set_header_property("data_offset", new_offset)
            self.set_header_property("num_variable_len_recs", len(value))
            self.data_provider.fileref.seek(0, 0)
            dat_part_1 = self.data_provider.fileref.read(self.header.header_size)
            self.data_provider.fileref.seek(old_offset, 0)
            dat_part_2 = self.data_provider.fileref.read(current_size - old_offset)
            # Manually Close:
            self.data_provider.close(flush=False)
            self.data_provider.open("w+b")
            self.data_provider.fileref.write(dat_part_1)
            for vlr in value:
                byte_string = vlr.to_byte_string()
                self.data_provider.fileref.write(byte_string)
            self.data_provider.fileref.write(b"\x00"*current_padding)
            self.data_provider.fileref.write(dat_part_2)
            self.data_provider.fileref.close()
            self.data_provider.open("r+b")
            self.data_provider.map()
            self.data_provider.point_map()
            self.populate_vlrs()
        elif self.mode == "w" and not self.has_point_records:

            self.set_header_property("num_variable_len_recs", len(value))
            if (self.data_provider._mmap.size() < self.header.header_size + sum([len(x) for x in value])):
                old_offset = self.header.header_size
                self.data_provider.fileref.seek(0, 0)
                dat_part_1 = self.data_provider.fileref.read(self.header.header_size)
                # Manually Close:
                self.data_provider.close(flush=False)
                self.data_provider.open("w+b")
                self.data_provider.fileref.write(dat_part_1)
                for vlr in value:
                    byte_string = vlr.to_byte_string()
                    self.data_provider.fileref.write(byte_string)

                self.data_provider.fileref.close()
                self.data_provider.open("r+b")

                self.data_provider.remap()
                new_offset = self.header.header_size + sum([len(x) for x in value])
                self.set_header_property("data_offset", new_offset)

            self.seek(self.header.header_size, rel = False)
            for vlr in value:
                self.data_provider._mmap.write(vlr.to_byte_string())
            self.populate_vlrs()
            return
        else:
            current_size = self.data_provider._mmap.size()
            current_padding = self.get_padding()
            old_offset = self.header.data_offset
            new_offset = current_padding + self.header.header_size + sum([len(x) for x in value])
            self.set_header_property("data_offset", new_offset)
            self.set_header_property("num_variable_len_recs", len(value))
            self.data_provider.fileref.seek(0, 0)
            dat_part_1 = self.data_provider.fileref.read(self.header.header_size)
            self.data_provider.fileref.seek(old_offset, 0)
            dat_part_2 = self.data_provider.fileref.read(current_size - old_offset)
            # Manually Close:
            self.data_provider.close(flush=False)
            self.data_provider.open("w+b")
            self.data_provider.fileref.write(dat_part_1)
            for vlr in value:
                byte_string = vlr.to_byte_string()
                self.data_provider.fileref.write(byte_string)
            self.data_provider.fileref.write(b"\x00"*current_padding)
            self.data_provider.fileref.write(dat_part_2)
            self.data_provider.fileref.close()
            self.data_provider.open("r+b")
            self.data_provider.map()
            self.data_provider.point_map()
            self.populate_vlrs()


    def set_padding(self, value):
        '''Set the padding between end of VLRs and beginning of point data'''
        if value < 0:
            raise laspy.util.LaspyException("New Padding Value Overwrites VLRs")
        if self.mode == "w":
            if not self.has_point_records:
                self.data_provider.fileref.seek(self.vlr_stop, 0)
                self.data_provider.fileref.write(b"\x00"*value)
                self.data_provider.remap()
                return
            else:
                raise laspy.util.LaspyException("Laspy does not yet support assignment of EVLRs for files which already contain point records.")
        elif self.mode == "rw":
            old_offset = self.header.data_offset
            self.set_header_property("data_offset",
                                            self.vlr_stop +  value)
            #self.header.data_offset = self.vlr_stop + value
            self.data_provider._mmap.flush()
            self.seek(0, rel=False)
            dat_part_1 = self.data_provider._mmap.read(self.vlr_stop)
            self.seek(old_offset, rel = False)
            dat_part_2 = self.data_provider._mmap.read(len(self.data_provider._mmap) - old_offset)
            self.data_provider.close()
            self.data_provider.open("w+b")
            self.data_provider.fileref.write(dat_part_1)
            self.data_provider.fileref.write(b"\x00"*value)
            self.data_provider.fileref.write(dat_part_2)
            self.data_provider.close()
            self.__init__(self.data_provider.filename, self.mode)
            return(len(self.data_provider._mmap))
        elif self.mode == "r+":
            pass
        else:
            raise(laspy.util.LaspyException("Must be in write mode to change padding."))
        return(len(self.data_provider._mmap))

    def pad_file_for_point_recs(self,num_recs):
        '''Pad the file with null bytes out to a calculated length based on
        the data given. This is usually a side effect of set_dimension being
        called for the first time on a file in write mode. '''

        bytes_to_pad = num_recs * self.point_format.rec_len
        self.header.point_records_count = num_recs
        if self.evlrs in [False, []]:
            #old_size = self.data_provider.filesize()
            old_size = self.header.data_offset
            self.data_provider._mmap.flush()
            self.data_provider.fileref.seek(old_size, 0)
            self.data_provider.fileref.write(b"\x00" * (bytes_to_pad))
            self.data_provider.fileref.flush()
            self.data_provider.remap(flush = False, point_map = True)
            # Write Phase complete, enter rw mode?
            self.padded = num_recs
            return
        else:
            d1 = self.data_provider._mmap[0:self.header.data_offset]
            d2 = self.data_provider._mmap[self.header.data_offset:self.data_provider._mmap.size()]
            self.data_provider.close()
            self.data_provider.open("w+b")
            self.data_provider.fileref.write(d1)
            self.data_provider.fileref.write(b"\x00"*(bytes_to_pad))
            self.data_provider.fileref.write(d2)
            self.data_provider.close()
            self.data_provider.remap(point_map = True)
            self.header.start_wavefm_data_rec += bytes_to_pad
            if self.header.version == "1.4":
                self.header.start_first_evlr += bytes_to_pad

    def define_new_dimension(self,name, data_type,description = ""):
        old_vlrs = self.vlrs
        if self.has_point_records or not self.mode == "w":
            raise laspy.util.LaspyException("New dimensions may be defined only for write mode files which do not yet possess point records.")
        eb_vlrs = [x for x in self.vlrs if x.type == 1]
        if self.header.version == "1.4":
            eb_evlrs =  [x for x in self.evlrs if x.type == 1]
            old_evlrs = self.evlrs
        else:
            eb_evlrs = []
            old_evlrs = []

        if ("extra_bytes" in self.point_format.lookup) and (len(eb_vlrs) != 0):
            raise laspy.util.LaspyException("Adding a dimension is ambiguous when there are already extra bytes in the point records, but no VLR describing them.")
        new_dimension = laspy.header.ExtraBytesStruct(name = name, data_type = data_type, description = description)
        if len(eb_vlrs) + len(eb_evlrs) > 1:
            raise laspy.util.LaspyException("Only one ExtraBytes VLR currently allowed.")
        elif len(eb_vlrs) + len(eb_evlrs) == 1:
            if len(eb_vlrs) == 1:
                extra_dimensions = eb_vlrs[0].extra_dimensions
                extra_dimensions.append(new_dimension)
                self.extra_dimensions = extra_dimensions
                new_pt_fmt = laspy.util.Format(self.point_format.fmt, extradims
                             = extra_dimensions)
                self.point_format = new_pt_fmt
                self.set_header_property("data_record_length", self.point_format.rec_len)

                eb_vlr_index = [x for x in range(len(self.vlrs)) if self.vlrs[x].type == 1][0]
                new_vlr = copy.copy(eb_vlrs[0])
                nvlrbs = new_dimension.to_byte_string()
                new_vlr.VLR_body += nvlrbs
                new_vlr.rec_len_after_header += len(nvlrbs)
                old_vlrs[eb_vlr_index] = new_vlr
                self.set_vlrs(old_vlrs)
                self.populate_vlrs()
            elif len(eb_evlrs) == 1:
                extra_dimensions = eb_evlrs[0].extra_dimensions
                extra_dimensions.append(new_dimension)
                self.extra_dimensions = extra_dimensions
                new_pt_fmt = laspy.util.Format(self.point_format.fmt, extradims
                             = extra_dimensions)
                self.point_format = new_pt_fmt
                self.set_header_property("data_record_length", self.point_format.rec_len)
                eb_evlr_index = [x for x in range(len(self.evlrs)) if self.evlrs[x].type == 1][0]
                new_vlr = copy.copy(eb_evlrs[0])
                nvlrbs = new_dimension.to_byte_string()
                new_vlr.VLR_body += nvlrbs
                new_vlr.rec_len_after_header += len(nvlrbs)
                old_vlrs[eb_evlr_index] = new_vlr
                self.set_evlrs(old_evlrs)
                self.populate_evlrs()
        else:
            # There are no current extra dimensions.
            new_vlr = laspy.header.VLR(user_id = "LASF_Spec", record_id = 4, VLR_body = new_dimension.to_byte_string())
            old_vlrs.append(new_vlr)
            self.extra_dimensions = [new_dimension]

            new_pt_fmt = laspy.util.Format(self.point_format.fmt, extradims = self.extra_dimensions)
            self.point_format = new_pt_fmt
            self.set_header_property("data_record_length", self.point_format.rec_len)
            self.set_vlrs(old_vlrs)
            self.populate_vlrs()

    def set_dimension(self, name,new_dim):
        '''Set a dimension (X,Y,Z etc) to the given value.'''
        #if not "__len__" in dir(new_dim):
        if isinstance(new_dim, GeneratorType):
            new_dim = list(new_dim)

        if not self.has_point_records:
            self.has_point_records = True
            self.set_header_property("point_records_count", len(new_dim))
            self.pad_file_for_point_recs(len(new_dim))


        ptrecs = self.get_pointrecordscount()
        if len(new_dim) != ptrecs:
            raise laspy.util.LaspyException("Error, new dimension length (%s) does not match"%str(len(new_dim)) + " the number of points (%s)" % str(ptrecs))
        try:
            spec = self.point_format.lookup[name]
            return(self._set_dimension(spec, new_dim))
        except KeyError:
            raise laspy.util.LaspyException("Dimension: " + str(name) +
                            " not found. Existing dimensions: " + ", ".join(self.point_format.lookup.keys()))

    def _set_dimension(self, spec, value):
        self.data_provider._pmap["point"][spec.name] = value
        return

    def _set_dimension_by_spec(self,new_dim,offs, fmt, length):
        '''Set a point dimension of appropriate offset format and length to new_dim'''
        if type(self.point_refs) == bool:
            self.build_point_refs()
        _mmap = self.data_provider._mmap
        packer = self.c_packers[fmt]
        i = 0
        for start in self.point_refs:
            _mmap[start+offs:start+offs+length] = packer.pack(new_dim[i])
            i += 1

        #idx = xrange(self.calc_point_recs)
        #starts = (self.point_refs[i] + offs for i in idx)
        #def f_set(x):
        #    i = starts.next()
        #    #self.seek(i, rel = False)
        #    #self.data_provider._mmap.write(pack(fmt, new_dim[x]))
        #    self.data_provider._mmap[i:i + length] = pack(fmt,new_dim[x])
        #map(f_set, idx)

        # Is this desireable
        #self.data_provider._mmap.flush()    def write_bytes(self, idx, bytes):
        return True

    def set_points(self, points):
        '''Set the point data for the file, using either a list of laspy.util.Point
        instances, or a numpy array of point data (as recieved from get_points).'''
        if isinstance(points, GeneratorType):
            points = list(points)
        if not self.has_point_records:
            self.has_point_records = True
            self.pad_file_for_point_recs(len(points))
        if isinstance(points[0], laspy.util.Point):
            self.data_provider._mmap[self.header.data_offset:self.data_provider._mmap.size()] = b"".join([x.pack() for x in points])
            self.data_provider.point_map()
        else:
             #self.data_provider._mmap[self.header.data_offset:self.data_provider._mmap.size()] = points.tostring()
             #self.data_provider._pmap["point"] = points["point"]
             self.data_provider._pmap[:] = points[:]
             #self.data_provider.point_map()
        #single_fmt = self.point_format.pt_fmt_long[1:]
        #big_fmt_string = "".join(["<", single_fmt*self.header.point_records_count])
        #out = []
        #(point.unpacked for point in points)
        #for i in points:
        #    out.extend(i.unpacked)
        #bytestr = pack(big_fmt_string, *out)
        #self.data_provider._mmap[self.header.data_offset:self.data_provider._mmap.size()] = bytestr

    def _set_raw_points(self, new_raw_points):
        if not self.has_point_records:
            self.has_point_records = True
            self.pad_file_for_point_recs(len(new_raw_points))
        '''Set a point dimension of appropriate name to new_dim'''
        ptrecs = self.get_pointrecordscount()
        if len(new_raw_points) != ptrecs:
            raise laspy.util.LaspyException("Error, new dimension length (%s) does not match"% str(len(new_raw_points))
                                            + " the number of points (%s)" % str(ptrecs))
        if type(self.point_refs) == bool:
            self.build_point_refs()
        idx = (xrange(len(self.point_refs)))
        def f(x):
            self.data_provider._mmap[self.point_refs[x]:self.point_refs[x]
                    + self.header.data_record_length] = new_raw_points[x]
        map(f, idx)
        self.data_provider.point_map()

    def _set_raw_datum(self, rec_offs, spec, val):
        '''Set a non dimension field with appropriate record type offset (0 for header)
        , appropriate spec object, and a new value. Uses raw bytes.'''
        self.data_provider._mmap[rec_offs+spec.offs:rec_offs+spec.offs +
                  spec.num*spec.length] = val
        return

    def _set_datum(self, rec_offs, dim, val):
        '''Set a non dimension field as with _set_raw_datum, but supply a formatted value'''

        if dim.num == 1:
            lb = rec_offs + dim.offs
            ub = lb + dim.length
            try:
                self.data_provider._mmap[lb:ub] = struct.pack(dim.fmt, val)
            except:
                self.data_provider._mmap[lb:ub] = struct.pack(dim.fmt, int(val))

            return

        try:
            dimlen = len(val)
        except(Exception):
            dimlen = 1

        if dim.num != dimlen:
            raise(laspy.util.LaspyException("Fields must be replaced with data of the same length. " +
                                str(dim.name) +" should be length " +
                                str(dim.num) +", received " + str(dimlen) ))
        def f(x):
            try:
                # because val is a bytes() object, val[x] yields an integer in python 3,
                # while val[x:x+1] returns bytes in both python 2 and 3
                q = val[x:x+1]
                outbyte = struct.pack(dim.fmt, q)
            except:
                outbyte = struct.pack(dim.fmt, int(val[x]))
            self.data_provider._mmap[(x*dim.length + rec_offs +
                    dim.offs):((x+1)*dim.length + rec_offs
                    + dim.offs)]=outbyte
        list(map(f, xrange(dim.num)))
        return

    def set_raw_header_property(self, name, value):
        '''Wrapper for _set_raw_datum, accpeting name of header property and raw byte value. '''
        try:
            spec = self.header_format.lookup[name]
        except(KeyError):
            raise(laspy.util.LaspyException("Header Dimension: " +
                  str(name) + " not found."))
        self._set_raw_datum(0, spec, value)

    def set_header_property(self, name, value):
        '''Wrapper for _set_datum, accepting name of header property and formatted value'''
        try:
            dim = self.header_format.lookup[name]
        except(KeyError):
            raise laspy.util.LaspyException("Header Dimension: " + str(name) + " not found.")
        if not dim.overwritable:
            raise(laspy.util.LaspyException("Field " + dim.name + " is not overwritable."))

        if dim.fmt == '<s':
            value = value.encode('ascii')
        self._set_datum(0, dim, value)
        self.header_changes.add(name)
        return

    def set_header(self, header):
        raise NotImplementedError

    def set_input_srs(self, srs):
        raise NotImplementedError

    def set_output_srs(self, srs):
        raise NotImplementedError

    ##  To Implement: Scale
    def set_x(self,X, scale = False):
        '''Wrapper for set_dimension("X", new_dimension)'''
        if not scale:
            self.set_dimension("X", X)
            return
        self.set_dimension("X", np.round((X - self.header.offset[0])/self.header.scale[0]))
        return

    def set_y(self,Y, scale = False):
        '''Wrapper for set_dimension("Y", new_dimension)'''
        if not scale:
            self.set_dimension("Y", Y)
            return
        self.set_dimension("Y", np.round((Y - self.header.offset[1])/self.header.scale[1]))
        return

    def set_z(self, Z, scale = False):
        '''Wrapper for set_dimension("Z", new_dimension)'''
        if not scale:
            self.set_dimension("Z", Z)
            return
        self.set_dimension("Z", np.round((Z-self.header.offset[2])/self.header.scale[2]))
        return

    def set_intensity(self, intensity):
        '''Wrapper for set_dimension("intensity", new_dimension)'''
        self.set_dimension("intensity", intensity)
        return

    def set_flag_byte(self, byte):
        '''Wrapper for set_dimension("flag_byte", new_dimension)'''
        self.set_dimension("flag_byte", byte)
        return

    # Utility Functions, refactor

    def bitpack(self, arrs, indices):
        '''Pack an array of integers into a byte based on idx
        for example bitpack((arr1, arr2), (0,3), (3,8)) packs the integers
        arr1 and arr2 into a byte, using the first three bits
        of arr1, and the last five bits of arr2.
        '''
        def keep_bits(arr, low, high):
            """ Keep only the bits on the interval [low, high) """
            return np.bitwise_and(np.bitwise_and(arr, 2**high - 1),
                                  ~(2**low - 1)).astype(np.uint8)

        first_bit_idx = 0 # Stack the bits from the beginning

        packed = np.zeros_like(arrs[0])
        for arr, (low, high) in zip(arrs, indices):
            if low > first_bit_idx:
                packed += np.right_shift(keep_bits(arr, low, high),
                                         low - first_bit_idx)
            else:
                packed += np.left_shift(keep_bits(arr, low, high),
                                        first_bit_idx - low)

            # First bit index should never be > 8 if we are
            # packing values to a byte
            first_bit_idx += high - low

            if first_bit_idx > 8:
                raise laspy.util.LaspyException("Invalid Data: Packed Length is Greater than allowed.")

        return list(packed)

    def raise_if_overflow(self, arr, maximum_packed_length):
        if np.any(np.array(arr) >= 2**maximum_packed_length):
            raise laspy.util.LaspyException("Invalid Data: Packed Length is Greater than allowed.")


    def set_return_num(self, num):
        '''Set the binary field return_num in the flag_byte'''
        self._header_current = False
        if self.header.data_format_id in (0,1,2,3,4,5):
            flag_byte = self.get_flag_byte()
            self.raise_if_overflow(num, 3)
            outByte = self.bitpack((num,flag_byte), ((0,3), (3,8)))
            self.set_dimension("flag_byte", outByte)
        elif self.header.data_format_id in (6,7,8,9,10):
            flag_byte = self.get_flag_byte()
            self.raise_if_overflow(num, 4)
            outByte = self.bitpack((num,flag_byte), ((0,4), (4,8)))
            self.set_dimension("flag_byte", outByte)
        return

    def set_num_returns(self, num):
        '''Set the binary field num_returns in the flag_byte'''
        self._header_current = False
        if self.header.data_format_id in (0,1,2,3,4,5):
            flag_byte = self.get_flag_byte()
            self.raise_if_overflow(num, 3)
            outByte = self.bitpack((flag_byte, num,flag_byte), ((0,3),(0,3),(6,8)))
            self.set_dimension("flag_byte", outByte)
        elif self.header.data_format_id in (6,7,8,9,10):
            flag_byte = self.get_flag_byte()
            self.raise_if_overflow(num, 4)
            outByte = self.bitpack((flag_byte, num), ((0,4), (0,4)))
            self.set_dimension("flag_byte", outByte)
        return

    def set_scanner_channel(self, value):
        if not self.header.data_format_id in (6,7,8,9,10):
            raise laspy.util.LaspyException("Scanner Channel not present for point format: " + str(self.header.data_format_id))
        raw_dim = self.get_raw_classification_flags()
        self.raise_if_overflow(value, 2)
        outByte = self.bitpack((raw_dim, value, raw_dim), ((0,4), (0,2), (6,8)))
        self.set_raw_classification_flags(outByte)

    def set_scan_dir_flag(self, flag):
        '''Set the binary field scan_dir_flag in the flag_byte'''
        if self.header.data_format_id in (0,1,2,3,4,5):
            flag_byte = self.get_flag_byte()
            self.raise_if_overflow(flag, 1)
            outByte = self.bitpack((flag_byte,flag,flag_byte),
                ((0,6),(0,1), (7,8)))
            self.set_dimension("flag_byte", outByte)
        elif self.header.data_format_id in (6,7,8,9,10):
            flag_byte = self.get_raw_classification_flags()
            self.raise_if_overflow(flag, 1)
            outByte = self.bitpack((flag_byte,flag,flag_byte),
                ((0,6),(0,1), (7,8)))
            self.set_dimension("classification_flags", outByte)
        return

    def set_edge_flight_line(self, line):
        '''Set the binary field edge_flight_line in the flag_byte'''
        if self.header.data_format_id in (0,1,2,3,4,5):
            raw_dim = self.get_flag_byte()
            self.raise_if_overflow(line, 1)
            outByte = self.bitpack((raw_dim, line), ((0,7), (0,1)))
            self.set_dimension("flag_byte", outByte)
        elif self.header.data_format_id in (6,7,8,9,10):
            raw_dim = self.get_raw_classification_flags()
            self.raise_if_overflow(line, 1)
            outByte = self.bitpack((raw_dim, line), ((0,7), (0,1)))
            self.set_dimension("classification_flags", outByte)
        return

    def set_classification_byte(self, value):
        self.set_dimension("classification_byte", value)

    def set_raw_classification_flags(self, value):
        self.set_dimension("classification_flags",value)

    def set_classification_flags(self, value):
        if not self.header.data_format_id in (6,7,8,9,10):
            self.set_classification(value)
            return
        rawDim = self.get_raw_classification_flags()
        self.raise_if_overflow(value, 4)
        outbyte = self.bitpack((value, rawDim), ((0,4), (4,8)))
        self.set_raw_classification_flags(outbyte)
        return

    def set_raw_classification(self, classification):
        '''Set the entire classification byte at once. This is faster than setting the binary fields individually,
        but care must be taken that the values mean what you think they do. '''
        self.set_dimension("raw_classification", classification)

    def set_classification(self, classification):
        '''Point Formats <6: Set the binary classification field inside the raw classification byte
           Point Formats >5: Set the classification byte.
        '''
        if self.header.data_format_id in (0,1,2,3,4,5):
            class_byte = self.get_raw_classification()
            self.raise_if_overflow(classification, 5)
            out_byte = self.bitpack((classification, class_byte), ((0,5), (5,8)))
            self.set_raw_classification(out_byte)
        elif self.header.data_format_id in (6,7,8,9,10):
            self.set_dimension("classification_byte", classification)
        return

    def set_synthetic(self, synthetic):
        '''Set the binary field synthetic inside the raw classification byte'''
        if self.header.data_format_id in (6,7,8,9,10):
            class_byte = self.get_raw_classification_flags()
            self.raise_if_overflow(synthetic, 1)
            out_byte = self.bitpack((synthetic, class_byte),
                                    ((0,1), (1,8)))
            self.set_raw_classification_flags(out_byte)
        else:
            class_byte = self.get_raw_classification()
            self.raise_if_overflow(synthetic, 1)
            out_byte = self.bitpack((class_byte, synthetic, class_byte),
                                   ((0,5), (0,1), (6,8)))
            self.set_dimension("raw_classification", out_byte)
        return

    def set_key_point(self, pt):
        '''Set the binary key_point field inside the raw classification byte'''
        if self.header.data_format_id in (6,7,8,9,10):
            class_byte = self.get_raw_classification_flags()
            self.raise_if_overflow(pt, 1)
            outbyte = self.bitpack((class_byte, pt, class_byte), ((0,1),(0,1), (2,8)))
            self.set_raw_classification_flags(outbyte)
        else:
            class_byte = self.get_raw_classification()
            self.raise_if_overflow(pt, 1)
            out_byte = self.bitpack((class_byte, pt, class_byte),
                                ((0,6),(0,1),(7,8)))
            self.set_dimension("raw_classification", out_byte)
        return

    def set_withheld(self, withheld):
        '''Set the binary field withheld inside the raw classification byte'''
        if self.header.data_format_id in (6,7,8,9,10):
            class_byte = self.get_raw_classification_flags()
            self.raise_if_overflow(withheld, 1)
            outbyte = self.bitpack((class_byte, withheld, class_byte), ((0,2),(0,1), (3,8)))
            self.set_raw_classification_flags(outbyte)
        else:
            class_byte = self.get_raw_classification()
            self.raise_if_overflow(withheld, 1)
            out_byte = self.bitpack((class_byte, withheld),
                                 ((0,7), (0,1)))
            self.set_dimension("raw_classification", out_byte)
        return

    def set_overlap(self, overlap):
        '''Set the binary field withheld inside the raw classification byte'''
        if self.header.data_format_id in (6,7,8,9,10):
            class_byte = self.get_raw_classification_flags()
            self.raise_if_overflow(overlap, 1)
            outbyte = self.bitpack((class_byte, overlap, class_byte), ((0,3),(0,1), (4,8)))
            self.set_raw_classification_flags(outbyte)
        else:
            raise laspy.util.LaspyException("Overlap only present in point formats > 5.")
        return


    def set_scan_angle_rank(self, rank):
        '''Wrapper for set_dimension("scan_angle_rank")'''
        self.set_dimension("scan_angle_rank", rank)
        return

    def set_scan_angle(self, rank):
        '''Wrapper for set_dimension("scan_angle")'''
        self.set_dimension("scan_angle", rank)
        return

    def set_user_data(self, data):
        '''Wrapper for set_dimension("user_data")'''
        self.set_dimension("user_data", data)
        return

    def set_pt_src_id(self, data):
        '''Wrapper for set_dimension("pt_src_id")'''
        self.set_dimension("pt_src_id", data)
        return

    def set_gps_time(self, data):
        '''Wrapper for set_dimension("gps_time")'''
        self.set_dimension("gps_time", data)

    def set_red(self, red):
        '''Wrapper for set_dimension("red")'''
        self.set_dimension("red", red)

    def set_green(self, green):
        '''Wrapper for set_dimension("green")'''
        self.set_dimension("green", green)

    def set_blue(self, blue):
        '''Wrapper for set_dimension("blue")'''
        self.set_dimension("blue", blue)

    def set_nir(self, value):
        self.set_dimension("nir", value)

    def set_wave_packet_desc_index(self, idx):
        '''Wrapper for set_dimension("wave_packet_desc_index") This is not currently functional,
        since addition of waveform data broke the numpy point map.'''
        self.set_dimension("wave_packet_desc_index", idx)

    def set_byte_offset_to_waveform_data(self, idx):
        '''Wrapper for set_dimension("byte_offset_to_waveform_data"), not currently functional,
        because addition of waveform data broke the numpy point map.'''
        self.set_dimension("byte_offset_to_waveform_data", idx)

    def set_waveform_packet_size(self, size):
        '''Wrapper for set_dimension("waveform_packet_size"), not currently functional, because
        addition of waveform data broke the numpy point map.'''
        self.set_dimension("waveform_packet_size", size)

    def set_return_point_waveform_loc(self, loc):
        '''Wrapper for set_dimension("return_point_waveform_loc"), not currently functional,
        because addition of waveform data broke the numpy point map.'''
        self.set_dimension("return_point_waveform_loc", loc)

    def set_x_t(self, x):
        '''Wrapper for set_dimension("x_t")'''
        self.set_dimension("x_t", x)

    def set_y_t(self, y):
        '''Wrapper for set_dimension("y_t")'''
        self.set_dimension("y_t", y)

    def set_z_t(self, z):
        '''Wrapper for set_dimension("z_t")'''
        self.set_dimension("z_t", z)

    def set_extra_bytes(self, extra_bytes):
        '''Wrapper for set_dimension("extra_bytes")'''
        if "extra_bytes" in self.point_format.lookup.keys():
            self.set_dimension("extra_bytes", extra_bytes)
        elif self.extra_dimensions != []:
            newmap = self.data_provider.get_point_map(self.naive_point_format)
            newmap["point"]["extra_bytes"] = extra_bytes
        else:
            raise laspy.util.LaspyException("Extra bytes not present in point format. Try creating a new file with an extended point record length.")




import laspy
import numpy as np
import argparse
import logging

class validate():
    def __init__(self):
        self.parse_args()
        self.clear_log()
        self.errors = 0
        self.tests = 0

    def parse_args(self):        
        parser = argparse.ArgumentParser(description="""Accept the path to a .LAS file, 
                                                    and print a list of point records 
                                                    with invalid (X,Y,Z) information.""")

        parser.add_argument("in_file", metavar="Input File", type = str, nargs=1, help = "Path to input file")
        parser.add_argument("--log", metavar="Log File", type = str, nargs=1, help ="Path to log file", default = "lasvalidate.log")
        parser.add_argument("--tol", metavar="Tolerance", type = float, nargs = 1, help = "Tolerance for header max/min vs actual max/min comparisons.", default = 0.01)
        self.args = parser.parse_args()

    def clear_log(self):
        tmp = open(self.args.log, "w")
        tmp.close()

    def bb(self,inFile, xval, yval, zval):
        X_invalid = np.logical_or((inFile.header.min[0] > xval), (inFile.header.max[0] < xval))
        Y_invalid = np.logical_or((inFile.header.min[1] > yval), (inFile.header.max[1] < yval))
        Z_invalid = np.logical_or((inFile.header.min[2] > zval), (inFile.header.max[2] < zval))

        bad_indices = np.where(np.logical_or(X_invalid, Y_invalid, Z_invalid))
        return(bad_indices)

    def test0(self, inFile):
        print("Test 0: Does this look like a LAS file?")
        if inFile.header.file_signature != "LASF":
            logging.info("Incorrect File Signature")
            print("... failed")
            self.errors += 1
        if self.errors == 0:
            print("...passed")
        self.tests += 1


    def test1(self, inFile):
        print("Test 1: Checking that all points fall inside header bounding box: ")
        bb = zip(["X", "Y", "Z"], inFile.header.min, inFile.header.max)
        print("... Header bounding box:")
        for i in bb:
            print("..." + str(i))

        bad_indices = self.bb(inFile, inFile.x, inFile.y, inFile.z)

        if len(bad_indices[0] == 0):
            print("... Header bounding box errors detected.")
            if len(bad_indices[0]) == len(inFile):
                new_bi = self.bb(inFile, inFile.X, inFile.Y, inFile.Z)[0]
                if len(new_bi) < len(inFile):
                    print("... Using unscaled values gives fewer invalid points: " + str(len(new_bi)) + ", header appears to use unscaled data.")
                    logging.info("Header appears to use unscaled data in max/min. This may follow spec, but is less common than using scaled data. ")
                else:
                    print("... Bounding box appears to be invalid.")
                    logging.info("Bounding box appears to be invalid.")
                    
            else:
                print("... printing bad indices to log: " + self.args.log)
                for i in bad_indices[0]:
                    logging.info("Point outside header bounding box: %i" % i)
            self.errors += 1
        else:
            print("... passed.")
        self.tests += 1

    def test2(self, inFile):
        print("Test 2: Checking that header bounding box is precise.")
        actual_max = [np.max(vec) for vec in [inFile.x, inFile.y, inFile.z]]
        actual_min = [np.min(vec) for vec in [inFile.x, inFile.y, inFile.z]]
        header_max = inFile.header.max 
        header_min = inFile.header.min
        max_diffs = [actual_max[i] - header_max[i] for i in range(3)]
        min_diffs = [actual_min[i] - header_min[i] for i in range(3)]
        err = 0
        for i in range(len(max_diffs)):
            if max_diffs[i] > self.args.tol:
                err += 1
                print("... " + ["X", "Y", "Z"][i] + " header max doesn't match actual max.")
                print("...    actual max: " + str(actual_max[i]) + ", header max: " + str(header_max[i]))
                logging.info(["X", "Y", "Z"][i] + " header max doesn't match actual max at tolerance %s." %str(self.args.tol))
        for i in range(len(min_diffs)):
            if min_diffs[i] > self.args.tol:
                err += 1
                print("... " + ["X", "Y", "Z"][i] + " header min doesn't match actual min.")
                print("...    actual min: " + str(actual_max[i]) + ", header min: " + str(header_max[i]))
                logging.info(["X", "Y", "Z"][i] + " header min doesn't match actual min at tolerance %s." %str(self.args.tol))
        if err == 0:
            print("... passed")
        else:
            self.errors += 1
            print("... failed")
        self.tests += 1

    def test3(self, inFile):
        print("Test 3: Checking that X and Y range values are reasonable.")
        X_range = np.max(inFile.x) - np.min(inFile.x)
        Y_range = np.max(inFile.y) - np.min(inFile.y)
        err = 0
        if X_range > 2000:
            err += 1
            print("... X range may be unreasonable: " + str(X_range))
            logging.info("X range may be unreasonable: " + str(X_range))
        if Y_range > 2000:
            err += 1
            print("... Y range may be unreasonable: " + str(Y_range))
            logging.info("Y range may be unreasonable: " + str(Y_range))
        if err == 0:
            print("... passed")
        else:
            print("... failed")
            self.errors += 1
        self.tests += 1

    def validate(self):
        print("Reading in file: " + self.args.in_file[0])
        inFile = laspy.file.File(self.args.in_file[0], mode = "r")
        self.test0(inFile)
        self.test1(inFile)
        self.test2(inFile)
        self.test3(inFile)
        print(str(self.errors) + " errors out of " + str(self.tests) + " tests")


def main():
    validator = validate()  
    logging.basicConfig(filename=validator.args.log,level=logging.DEBUG)
    validator.validate()





import laspy
import argparse

try:
    xrange
except NameError:
    xrange = range

class lasverify():
    def __init__(self):
        self.parse_args()
        self.verify()
        ## Setup argparse to accept command line arguments
    def parse_args(self):
        parser = argparse.ArgumentParser(
                description='''Compare LAS files file_1 and file_2 by dimension, header
                       field, vlr and evlr.''')
        parser.add_argument('file_1', metavar='file_1', type=str, nargs='+',
                            help='LAS file 1 path.')
        parser.add_argument('file_2', metavar='file_2', type=str, nargs='+',
                            help='LAS file 2 path.')

        parser.add_argument('-b', type=bool,help='''Attempt to compare incompatable 
                sub-byte sized data fields if present? (slow)''', default=False)

        self.args = parser.parse_args()

    def verify(self):
    ## Set global varibles for convenience.
        file_1 = self.args.file_1[0]
        file_2 = self.args.file_2[0]
        PRESERVE = self.args.b 

    ## Try to open both files in read mode.
        try:
            inFile1 = laspy.file.File.File(file_1,mode= "r")
            inFile2 = laspy.file.File.File(file_2,mode= "r")
        except Exception as error:
            print("Error reading in files:")
            print(error)
            quit()

    ## Set global flag to indicate whether we need to look at incompatable 
    ## bit field bytes. 
        SUB_BYTE_COMPATABLE = ((inFile1.header.data_format_id <= 5) == 
                            (inFile2.header.data_format_id <= 5)) 

    ## Warn the user if they chose not to check incompatable point formats.
        if (not SUB_BYTE_COMPATABLE) and (not PRESERVE):
            print("""WARNING: Point formats %i and %i have mismatched sub-byte fields. 
                    The default behavior in this case is to ignore these fields during 
                    the file_verify procedure. If you want laspy to attempt to match up
                    sub-byte data between these two formats, specify -b=True 
                    (this might take some time depending on the size of the file)""")

        def print_title(string):
            print("#"*65)
            print("#  " + string + " "*(65 - (len(string) + 4)) + "#")
            print("#"*65)

    ## Define convenience function to try to compare point dimensions. 
        def f(x):
            try:
                return(1*(list(inFile1.reader.get_dimension(x)) == list(inFile2.reader.get_dimension(x))))
            except:
                outstr = "Dimension: %s" %x
                outstr += " "*(50-len(outstr))
                if not (x in inFile1.point_format.lookup) and (x in inFile2.point_format.lookup):
                    print(outstr + "not present in file_1.")
                elif not (x in inFile2.point_format.lookup) and (x in inFile1.point_format.lookup):
                    print(outstr + "not present in file_2.")
                else:
                    print("There was an error comparing dimension: %s" + str(x))
                return(2)

    ## Define convenience function to try to compare header fields. 
        def g(x):
            try:
                return((1*(inFile1.reader.get_header_property(x) == inFile2.reader.get_header_property(x))))
            except:
                outstr = "Header Property: %s" % x
                outstr += (" "*(50-len(outstr)))
                if not (x in inFile1.header.header_format.lookup) and (x in inFile2.header.header_format.lookup):
                    print(outstr + "not present in file_1.")
                elif not (x in inFile2.header.header_format.lookup) and (x in inFile1.header.header_format.lookup):
                    print(outstr + "not present in file_2.")
                else:
                    print("There was an error while comparing header property: %s" + str(x))
                return(2)

    ## Build union fo header fields, then check them.
        print_title("Checking Headers")
        try:
            passed = 0
            failed = 0
            header_props = set()
            for item in inFile1.reader.header_format.specs:
                header_props.add(item.name)
            for item in inFile2.reader.header_format.specs:
                header_props.add(item.name)
            for item in header_props:
                outstr = "Header Property: %s" % item
                outstr += " "*(50-len(outstr))
                result = g(item)
                if result == 1:
                    print(outstr + "identical")
                    passed += 1
                elif not result == 2:
                    print(outstr + "different")
                    print("   File 1: " + str(inFile1.reader.get_header_property(item)))
                    print("   File 2: " + str(inFile2.reader.get_header_property(item)))
                    failed += 1
            print("%i of %i header fields match." % (passed, passed + failed))
        except:
            print("There was an error while comparing headers. ")

    ## Check VLRs
        print_title("Checking VLRs")

        try:
            if len(inFile1.header.vlrs) != len(inFile2.header.vlrs):
                print("Number of VLRs differs: file_1 has %i and file_2 has %i. Comparing where possible..."
                        % (len(inFile1.header.vlrs), len(inFile2.header.vlrs)))
            for i in xrange(min(len(inFile1.header.vlrs), len(inFile2.header.vlrs))):
                outstr = "VLR Record: %i" %i
                outstr += " "*(50 - len(outstr))
                vlr1 = inFile1.header.vlrs[i].to_byte_string()
                vlr2 = inFile2.header.vlrs[i].to_byte_string()
                if vlr1 == vlr2:
                    print(outstr + "identical")
                else:
                    print(outstr + "different")
        except:
            print("There was a problem comparing VLRs")

    ## Check EVLRs
        print_title("Checking EVLRs")
        try:
            if len(inFile1.header.evlrs) != len(inFile2.header.evlrs):
                print("Number of VLRs differs: file_1 has %i and file_2 has %i. Comparing where possible..."
                        % (len(inFile1.header.evlrs), len(inFile2.header.evlrs)))

            for i in xrange(min(len(inFile1.header.evlrs), len(inFile2.header.evlrs))):
                outstr = "EVLR Record: %i" %i
                outstr += " "*(50 - len(outstr))
                vlr1 = inFile1.header.evlrs[i].to_byte_string()
                vlr2 = inFile1.header.evlrs[i].to_byte_string()
                if vlr1 == vlr2:
                    print(outstr + "identical")
                else:
                    print(outstr + "different")
        except:
            print("There was a problem comparing EVLRs")

    ## Build a union of dimensions in each file, then compare them.
        dims = set()
        for item in inFile1.reader.point_format.lookup.keys():
            dims.add(item)
        for item in inFile2.reader.point_format.lookup.keys():
            dims.add(item)
        print_title("Checking Dimensions")
        passed = 0
        failed = 0
        for dim in dims:
            if not SUB_BYTE_COMPATABLE and dim in ("raw_classification","classification_flags", "classification_byte", "flag_byte"):
                continue
            outstr = "Dimension: %s" % dim 
            outstr += " "*(50-len(outstr))
            result = f(dim)
            if result == 1:
                passed += 1
                print(outstr +  "identical")
            elif result != 2:
                failed += 1
                print(outstr + "different")

        def print_sb(string, result):
            outstr  ="Sub-byte field: " + string
            outstr += " "*(50-len(outstr))
            if result:
                print(outstr + "identical")
                return(1)
            else:
                print(outstr + "different")
                return(0)

    ## If we need to, compare valid sub-byte fields
        sb_total = 0
        if not SUB_BYTE_COMPATABLE and PRESERVE:
            print("Comparing sub-byte fields (this might take awhile)")
            sb_total += print_sb("classification", all(inFile1.classification == inFile2.classification)) 
            sb_total += print_sb("return_num", all(inFile1.return_num == inFile2.return_num))
            sb_total += print_sb("num_returns", all(inFile1.num_returns == inFile2.num_returns))
            sb_total += print_sb("scan_dir_flag", all(inFile1.scan_dir_flag == inFile2.scan_dir_flag))
            sb_total += print_sb("edge_flight_line", all(inFile1.edge_flight_line == inFile2.edge_flight_line))
            sb_total += print_sb("synthetic", all(inFile1.synthetic == inFile2.synthetic))
            sb_total += print_sb("withheld", all(inFile1.withheld == inFile2.withheld))
            sb_total += print_sb("key_point", all(inFile1.key_point == inFile2.key_point))


    ## Print summary
        print(str(passed) + " identical dimensions out of " + str(passed + failed))
        if not SUB_BYTE_COMPATABLE and PRESERVE:
            print("%i identical sub-byte fields out of %i" %(sb_total, 8))

        inFile1.close()
        inFile2.close()

def main():
    vf = lasverify()
    print("Verify Complete")
